from __future__ import unicode_literals
from future.builtins import str, bytes, object

import uuid

from logging import getLogger

log = getLogger(__name__)


def make_key(*args):
    """Create the key concatening all args with `:`."""
    return u":".join(str(arg) for arg in args)


def unique_key(connection):
    """
    Generate a unique keyname that does not exists is the connection
    keyspace.
    """
    while 1:
        key = str(uuid.uuid4().hex)
        if not connection.exists(key):
            break
    return key


def normalize(value):
    """
    Simple method to always have the same kind of value
    """
    if value and isinstance(value, bytes):
        value = value.decode('utf-8')
    return value


class cached_property(object):
    """
    Decorator that converts a method with a single self argument into a
    property cached on the instance.
    Optional ``name`` argument allows you to make cached properties of other
    methods. (e.g.  url = cached_property(get_absolute_url, name='url') )

    From https://github.com/django/django/blob/27793431cf21a82809c0c39a7c0188a2d83bf475/django/utils/functional.py#L15
    """
    def __init__(self, func, name=None):
        self.func = func
        self.__doc__ = getattr(func, '__doc__')
        self.name = name or func.__name__

    def __get__(self, instance, cls=None):
        if instance is None:
            return self
        res = instance.__dict__[self.name] = self.func(instance)
        return res



# -*- coding:utf-8 -*-
from __future__ import unicode_literals
from future.utils import iteritems, iterkeys
from future.builtins import str
from future.builtins import zip
from future.utils import with_metaclass

from inspect import isclass
from logging import getLogger
from copy import copy

from redis import VERSION as redispy_version
from redis.exceptions import RedisError

from limpyd.database import Lock
from limpyd.utils import cached_property, make_key, normalize
from limpyd.exceptions import *

log = getLogger(__name__)

__all__ = [
    'InstanceHashField',
    'RedisField',
    'RedisProxyCommand',
    'MetaRedisProxy',
    'SortedSetField',
    'StringField',
    'ListField',
    'SetField',
    'PKField',
    'AutoPKField',
    'HashField',
]


class MetaRedisProxy(type):
    """
    This metaclass create the class normally, then takes a list of redis
    commands found in the "available_*" class attributes, and for each one
    create the corresponding method if it not exists yet. Created methods simply
    call _call_command.
    """

    @staticmethod
    def class_can_have_commands(klass):
        """
        Return False if the given class inherits from RedisProxyCommand, which
        indicates that it can handle its own sets of available commands.
        If not a subclass of RedisProxyCommand, (for example "object", to create
        a simple mixin), returns False.
        """
        try:
            return issubclass(klass, RedisProxyCommand)
        except NameError:
            # We pass here if we are workong on RedisProxyCommand itself, which
            # is not yet defined in this case
            return False

    def __new__(mcs, name, base, dct):
        """
        Create methods for all redis commands available for the given class.
        """
        it = super(MetaRedisProxy, mcs).__new__(mcs, name, base, dct)

        # It the class we are working on is not aimed to directly have its own
        # commands defined, we don't try to manage them.
        # It's needed to use mixins, which must be based on `object`.
        # See contrib.related.*RelatedFieldMixin to see an example of such a
        # mixin

        if any([mcs.class_can_have_commands(one_base) for one_base in base]):

            # make sure we have a set for each list of type of command
            for attr in ('available_getters', 'available_modifiers', ):
                setattr(it, attr, set(getattr(it, attr, ())))

            # add simplest set: getters, modidiers, all
            it.available_commands = it.available_getters.union(it.available_modifiers)

            # create a method for each command
            for command_name in it.available_commands:
                if not hasattr(it, command_name):
                    setattr(it, command_name, it._make_command_method(command_name))

        return it


class RedisProxyCommand(with_metaclass(MetaRedisProxy)):

    @classmethod
    def _make_command_method(cls, command_name):
        """
        Return a function which call _call_command for the given name.
        Used to bind redis commands to our own calls
        """
        def func(self, *args, **kwargs):
            return self._call_command(command_name, *args, **kwargs)
        return func

    def _call_command(self, name, *args, **kwargs):
        """
        Check if the command to be executed is a modifier, to connect the object.
        Then call _traverse_command.
        """
        obj = getattr(self, '_instance', self)  # _instance if a field, self if an instance

        # The object may not be already connected, so if we want to update a
        # field, connect it before.
        # If the object as no PK yet, let the object create itself
        if name in self.available_modifiers and obj._pk and not obj.connected:
            obj.connect()

        # Give priority to a "_call_{commmand}" method
        meth = getattr(self, '_call_%s' % name, self._traverse_command)
        return meth(name, *args, **kwargs)

    def _traverse_command(self, name, *args, **kwargs):
        """
        Add the key to the args and call the Redis command.
        """
        if not name in self.available_commands:
            raise AttributeError("%s is not an available command for %s" %
                                 (name, self.__class__.__name__))
        attr = getattr(self.connection, "%s" % name)
        key = self.key
        log.debug(u"Requesting %s with key %s and args %s" % (name, key, args))
        result = attr(key, *args, **kwargs)
        result = self.post_command(
            sender=self,
            name=name,
            result=result,
            args=args,
            kwargs=kwargs
        )
        return result

    def post_command(self, sender, name, result, args, kwargs):
        """
        Call after we got the result of a redis command.
        By default, does nothing, but must return a value.
        """
        return result

    @classmethod
    def get_connection(cls):
        """
        Return the connection from the database
        """
        return cls.database.connection

    @property
    def connection(self):
        """
        A simple property on the instance that return the connection stored on
        the class
        """
        return self.get_connection()


class RedisField(RedisProxyCommand):
    """
    Base class for all fields using redis data structures.
    """
    # The "_commands_to_proxy" dict take redis commands as keys, and proxy
    # method names as values. These proxy_methods must take the real command
    # name in first parameter, and a *args+**kwargs to pass needed values.
    # Their goal is to simplify management of values to index/deindex for simple
    # redis commands.
    _commands_to_proxy = {}

    _creation_order = 0  # internal class counter to keep fields ordered

    proxy_setter = None
    unique = False
    _copy_conf = {
        'args': [],
        'kwargs': ['lockable', 'default', 'indexable', 'unique', ('indexes', 'index_classes')],
        'attrs': ['name', '_instance', '_model']
    }
    _unique_supported = True
    _field_parts = 1
    default_indexes = None

    def __init__(self, *args, **kwargs):
        """
        Manage all field attributes
        """
        self.lockable = kwargs.get('lockable', True)
        if "default" in kwargs:
            self.default = kwargs["default"]

        self.indexable = kwargs.get("indexable", False)
        self.unique = kwargs.get("unique", False)
        if self.unique:
            if not self._unique_supported:
                raise ImplementationError('%s field cannot be unique' % self.__class__.__name__)
            if hasattr(self, "default"):
                raise ImplementationError('Cannot set "default" and "unique" together!')
            self.indexable = True

        self.index_classes = kwargs.get('indexes', [])
        if self.index_classes:
            if not self.indexable:
                raise ImplementationError('Cannot pass indexes if not indexable')

        # keep fields ordered
        self._creation_order = RedisField._creation_order
        RedisField._creation_order += 1

    def get_default_indexes(self):
        if self.__class__.default_indexes is not None:
            return self.__class__.default_indexes
        if hasattr(self, '_model'):
            return self._model.get_default_indexes()

    def proxy_get(self):
        """
        A helper to easily call the proxy_getter of the field
        """
        getter = getattr(self, self.proxy_getter)
        return getter()

    def proxy_set(self, value):
        """
        A helper to easily call the proxy_setter of the field
        """
        setter = getattr(self, self.proxy_setter)
        if isinstance(value, (list, tuple, set)):
            result = setter(*value)
        elif isinstance(value, dict):
            result = setter(**value)
        else:
            result = setter(value)
        return result

    @property
    def key(self):
        """
        A property to return the key used in redis for the current field.
        """
        return self.make_key(
            self._instance._name,
            self._instance.pk.get(),
            self.name,
        )

    @property
    def database(self):
        """
        A simple shortcut to access the database property of the field's instance
        """
        if not self._model:
            raise TypeError('A field cannot use a database if not linked to a model')
        return self._model.database

    @property
    def sort_wildcard(self):
        """
        Key used to sort models on this field.
        """
        return self.make_key(
            self._model._name,
            "*",
            self.name,
        )

    @property
    def connection(self):
        """
        A simple shortcut to get the connections of the field's instance's model
        """
        if not self._model:
            raise TypeError('A field cannot use a connection if not linked to a model')
        return self._model.get_connection()

    def __copy__(self):
        """
        In the RedisModel metaclass and constructor, we need to copy the fields
        to new ones. It can be done via the copy function of the copy module.
        This __copy__ method handles the copy by creating a new field with same
        attributes, without ignoring private attributes.
        Configuration of args and kwargs to pass to the constructor, and
        attributes to copy is done in the _copy_conf attribute of the class, a
        dict with 3 entries:
          - args: list of attributes names to pass as *args to the constructor
          - kwargs: list of attributes names to pass as **kwargs to the
                    constructor. If a tuple is used instead of a simple string
                    in the list, its first entry will be the kwarg name, and
                    the second the name of the attribute to copy
          - attrs: list of attributes names to copy (with "=") from the old
                   object to the new one
        """
        # prepare unnamed arguments
        args = [getattr(self, arg) for arg in self._copy_conf['args']]

        # prepare named arguments
        kwargs = {}
        for arg in self._copy_conf['kwargs']:
            # if arg is a tuple, the first entry will be the named kwargs, and
            # the second will be the name of the attribute to copy
            name = arg
            if isinstance(arg, tuple):
                name, arg = arg
            if hasattr(self, arg):
                kwargs[name] = getattr(self, arg)

        # create the new instance
        new_copy = self.__class__(*args, **kwargs)

        # then copy attributes
        for attr_name in self._copy_conf['attrs']:
            if hasattr(self, attr_name):
                setattr(new_copy, attr_name, getattr(self, attr_name))

        return new_copy

    def make_key(self, *args):
        """
        Simple shortcut to the make_key global function to create a redis key
        based on all given arguments.
        """
        return make_key(*args)

    def delete(self):
        """
        Delete the field from redis.
        """
        return self._call_command('delete')

    def post_command(self, sender, name, result, args, kwargs):
        """
        Call after we got the result of a redis command.
        By default, let the instance manage the post_modify signal
        """
        return self._instance.post_command(
                   sender=self,
                   name=name,
                   result=result,
                   args=args,
                   kwargs=kwargs
               )

    def exists(self):
        """
        Call the exists command to check if the redis key exists for the current
        field
        """
        try:
            key = self.key
        except DoesNotExist:
            """
            If the object doesn't exists anymore, its PK is deleted, so the
            "self.key" call will raise a DoesNotExist exception. We catch it
            to return False, as the field doesn't exists too.
            """
            return False
        else:
            return self.connection.exists(key)

    @cached_property
    def _indexes(self):
        """Instantiate the indexes only when asked

        Returns
        -------
        list
            An empty list if the field is not indexable, else a list of all indexes
            tied to the field.
            If no indexes where passed when creating the field, the default indexes
            from the field/model/database will be used.
            If still no index classes, it will raise

        Raises
        ------
        ImplementationError
            If no index classes available for this field

        """
        if not self.indexable:
            return []
        if not self.index_classes:
            self.index_classes = self.get_default_indexes()[::1]
        if not self.index_classes:
            raise ImplementationError('%s field is indexable but has no indexes attached' %
                                      self.__class__.__name__)

        return [index_class(field=self) for index_class in self.index_classes]

    def has_index(self, index):
        """Tells if the field have an index matching the current one

        Parameters
        -----------
        index: type or BaseIndex
            It could be an index instance, or an index class

        Returns
        -------
        bool
            Will be ``True`` if the current field has an index that is an instance
            of the given index or of the class of the given index

        """
        klass = index if isclass(index) else index.__class__
        for one_index in self._indexes:
            if isinstance(one_index, klass):
                return True
        return False

    def _attach_to_model(self, model):
        """
        Attach the current field to a model. Can be overriden to do something
        when a model is set
        """
        self._model = model

    def _attach_to_instance(self, instance):
        """
        Attach the current field to an instance of a model. Can be overriden to
        do something when an instance is set
        """
        self._instance = instance
        self.lockable = self.lockable and instance.lockable

    @property
    def attached_to_model(self):
        """Tells if the current field is the one attached to the model, not instance"""
        try:
            if not bool(self._model):
                return False
        except AttributeError:
            return False
        else:
            try:
                return not bool(self._instance)
            except AttributeError:
                return True

    def _call_command(self, name, *args, **kwargs):
        """
        Add lock management and call parent.
        """
        meth = super(RedisField, self)._call_command
        if self.indexable and name in self.available_modifiers:
            with FieldLock(self):
                try:
                    result = meth(name, *args, **kwargs)
                except:
                    self._rollback_indexes()
                    raise
                else:
                    return result
                finally:
                    self._reset_indexes_caches()
        else:
            return meth(name, *args, **kwargs)

    def _rollback_indexes(self):
        """
        Restore the index in its previous status, using deindexed/indexed values
        temporarily stored.
        """
        for index in self._indexes:
            index._rollback()

    def _reset_indexes_caches(self):
        """
        Reset attributes used to store deindexed/indexed values, used to
        rollback the index when something failed.
        """
        for index in self._indexes:
            index._reset_cache()

    def index(self, value=None, only_index=None):
        """
        Handle field index process.
        """
        assert self.indexable, "Field not indexable"
        assert not only_index or self.has_index(only_index), "Invalid index"
        if only_index:
            only_index = only_index if isclass(only_index) else only_index.__class__

        if value is None:
            value = self.proxy_get()

        if value is not None:
            needs_to_check_uniqueness = bool(self.unique)

            for index in self._indexes:
                if only_index and not isinstance(index, only_index):
                    continue

                index.add(value, check_uniqueness=needs_to_check_uniqueness and index.handle_uniqueness)

                if needs_to_check_uniqueness and index.handle_uniqueness:
                    # uniqueness check is done for this value
                    needs_to_check_uniqueness = False

    def deindex(self, value=None, only_index=None):
        """
        Run process of deindexing field value(s).
        """
        assert self.indexable, "Field not indexable"
        assert not only_index or self.has_index(only_index), "Invalid index"
        if only_index:
            only_index = only_index if isclass(only_index) else only_index.__class__

        if value is None:
            value = self.proxy_get()

        if value is not None:
            for index in self._indexes:
                if only_index and not isinstance(index, only_index):
                    continue
                index.remove(value)

    def clear_indexes(self, chunk_size=1000, aggressive=False, index_class=None):
        """Clear all indexes tied to this field

        Parameters
        ----------
        chunk_size: int
            Default to 1000, it's the number of instances to load at once if not in aggressive mode.
        aggressive: bool
            Default to ``False``. When ``False``, the actual collection of instances will
            be ran through to deindex all the values.
            But when ``True``, the database keys will be scanned to find keys that matches the
            pattern of the keys used by the indexes. This is a lot faster and may find forgotten keys.
            But may also find keys not related to the index.
            Should be set to ``True`` if you are not sure about the already indexed values.
        index_class: type
            Allow to clear only index(es) for this index class instead of all indexes.

        Raises
        ------
        AssertionError
            If called from an instance field. It must be called from the model field
            Also raised if the field is not indexable

        Examples
        --------

        >>> MyModel.get_field('myfield').clear_indexes()
        >>> MyModel.get_field('myfield').clear_indexes(index_class=MyIndex)

        """
        assert self.indexable, "Field not indexable"
        assert self.attached_to_model, \
            '`rebuild_indexes` can only be called on a field attached to the model'

        for index in self._indexes:
            if index_class and not isinstance(index, index_class):
                continue
            index.clear(chunk_size=chunk_size, aggressive=aggressive)

    def rebuild_indexes(self, chunk_size=1000, aggressive_clear=False, index_class=None):
        """Rebuild all indexes tied to this field

        Parameters
        ----------
        chunk_size: int
            Default to 1000, it's the number of instances to load at once.
        aggressive_clear: bool
            Will be passed to the `aggressive` argument of the `clear_indexes` method.
            If `False`, all values will be normally deindexed. If `True`, the work
            will be done at low level, scanning for keys that may match the ones used by the indexes
        index_class: type
            Allow to build only index(es) for this index class instead of all indexes.

        Raises
        ------
        AssertionError
            If called from an instance field. It must be called from the model field
            Also raised if the field is not indexable

        Examples
        --------

        >>> MyModel.get_field('myfield').rebuild_indexes()
        >>> MyModel.get_field('myfield').clear_indexes(index_class=MyIndex)


        """
        assert self.indexable, "Field not indexable"
        assert self.attached_to_model, \
            '`rebuild_indexes` can only be called on a field attached to the model'

        for index in self._indexes:
            if index_class and not isinstance(index, index_class):
                continue
            index.rebuild(chunk_size=chunk_size, aggressive_clear=aggressive_clear)

    def get_unique_index(self):
        assert self.unique, "Field not unique"

        try:
            return [index for index in self._indexes if index.handle_uniqueness][0]
        except IndexError:
            raise ImplementationError(
                'Field %s.%s is unique but has no indexes capable of handling uniqueness' %
                (self._model.__name__, self.name)
            )

    def check_uniqueness(self, value):
        if not self.unique:
            return
        if value is not None:
            self.get_unique_index().check_uniqueness(value)

    def from_python(self, value):
        """
        Coerce a value before using it in Redis.
        """
        return normalize(value)

    def _reset(self, command, *args, **kwargs):
        """
        Shortcut for commands that reset values of the field.
        All will be deindexed and reindexed.
        """
        if self.indexable:
            self.deindex()
        result = self._traverse_command(command, *args, **kwargs)
        if self.indexable:
            self.index()
        return result

    def _reindex_from_result(self, command, *args, **kwargs):
        """
        Same as _reset, but uses Redis return value to reindex, to
        save one query.
        """
        if self.indexable:
            self.deindex()
        result = self._traverse_command(command, *args, **kwargs)
        if self.indexable and result is not None:
            self.index(result)
        return result

    def _del(self, command, *args, **kwargs):
        """
        Shortcut for commands that remove all values of the field.
        All will be deindexed.
        """
        if self.indexable:
            self.deindex()
        return self._traverse_command(command, *args, **kwargs)
    _call_delete = _del


class SingleValueField(RedisField):
    """
    A simple parent class for StringField, InstanceHashField and PKField, all field
    types handling a single value.
    """

    def _call_set(self, command, value, *args, **kwargs):
        """
        Helper for commands that only set a value to the field.
        """
        if self.indexable:
            current = self.proxy_get()
            if normalize(current) != normalize(value):
                if current is not None:
                    self.deindex(current)
                if value is not None:
                    self.index(value)
        return self._traverse_command(command, value, *args, **kwargs)


class StringField(SingleValueField):

    proxy_getter = "get"
    proxy_setter = "set"

    available_getters = ('get', 'getbit', 'getrange', 'strlen', 'bitcount', )
    available_modifiers = ('delete', 'getset', 'set', 'append', 'decr',
                           'incr', 'incrbyfloat', 'setbit', 'setnx',
                           'setrange', )

    _call_getset = SingleValueField._call_set
    _call_append = _call_setrange = _call_setbit = SingleValueField._reset
    _call_decr = SingleValueField._reindex_from_result
    _call_incr = SingleValueField._reindex_from_result
    _call_incrbyfloat = SingleValueField._reindex_from_result

    def _call_setnx(self, command, value):
        """
        Index only if value has been set.
        """
        result = self._traverse_command(command, value)
        if self.indexable and value is not None and result:
            self.index(value)
        return result


class MultiValuesField(RedisField):
    """
    It's a base class for SetField, SortedSetField and ListField, to manage
    indexes when their constructor got the param "indexable" set to True.
    Indexes need more work than for simple RedisField as we have here many
    values in each field.
    A naive implementation is to simply deindex all existing values, call the
    wanted redis command, then reindex all.
    When possible, each commands of each impacted fields are done by catching
    values to index/deindex, to only do this work for needed values.
    It's the case for almost all defined commands, except ones where bulk
    removing is done, as zremrange* for sorted set and ltrim (and lrem in some
    cases) for lists (for these commands, the naive algorithm defined above is
    used, so use them carefully).
    See the _traverse_command method below to know how values to index/deindex
    are defined.
    """

    scannable = False

    def _add(self, command, *args, **kwargs):
        """
        Shortcut for commands that only add values to the field.
        Added values will be indexed.
        """
        if self.indexable:
            self.index(args)
        return self._traverse_command(command, *args, **kwargs)

    def _rem(self, command, *args, **kwargs):
        """
        Shortcut for commands that only remove values from the field.
        Removed values will be deindexed.
        """
        if self.indexable:
            self.deindex(args)
        return self._traverse_command(command, *args, **kwargs)

    def _pop(self, command, *args, **kwargs):
        """
        Shortcut for commands that pop a value from the field, returning it while
        removing it.
        The returned value will be deindexed
        """
        result = self._traverse_command(command, *args, **kwargs)
        if self.indexable:
            self.deindex([result])
        return result

    def index(self, values=None, only_index=None):
        """
        Index all values stored in the field, or only given ones if any.
        """
        assert self.indexable, "Field not indexable"
        assert not only_index or self.has_index(only_index), "Invalid index"
        if only_index:
            only_index = only_index if isclass(only_index) else only_index.__class__

        if values is None:
            values = self.proxy_get()

        for value in values:
            if value is not None:
                needs_to_check_uniqueness = bool(self.unique)

                for index in self._indexes:
                    if only_index and not isinstance(index, only_index):
                        continue

                    index.add(value, check_uniqueness=needs_to_check_uniqueness and index.handle_uniqueness)

                    if needs_to_check_uniqueness and index.handle_uniqueness:
                        # uniqueness check is done for this value
                        needs_to_check_uniqueness = False

    def deindex(self, values=None, only_index=None):
        """
        Deindex all values stored in the field, or only given ones if any.
        """
        assert self.indexable, "Field not indexable"
        assert not only_index or self.has_index(only_index), "Invalid index"
        if only_index:
            only_index = only_index if isclass(only_index) else only_index.__class__

        if not values:
            values = self.proxy_get()

        for value in values:
            if value is not None:
                for index in self._indexes:
                    if only_index and not isinstance(index, only_index):
                        continue
                    index.remove(value)

    def check_uniqueness(self, values):
        if not self.unique:
            return
        for value in values:
            if value is not None:
                self.get_unique_index().check_uniqueness(value)

    if redispy_version >= (2, 10, 0):

        def _scan(self, command, match=None, count=None):
            assert self.scannable, 'Field is not scannable'
            assert not command.endswith('_iter'), 'Use %s instead of %s' % (command[:-5], command)

            return self._traverse_command(command + '_iter', match=match, count=count)

    else:
        # no *scan_iter in redis-py < 2.10

        def _scan(self, command, match=None, count=None):
            assert self.scannable, 'Field is not scannable'

            cursor = 0
            while True:
                cursor, data = self._traverse_command(command, cursor, match=match, count=count)
                for item in (data.items() if hasattr(data, 'items') else data):
                    yield item
                if not cursor or cursor == '0':
                    break


class SortedSetField(MultiValuesField):
    """
    A field with values stored in a sorted set.
    If the indexable argument is set to True on the constructor, all stored
    values will be indexed. But when using zremrange* commands, all content will
    be deindexed and then reindexed as we have no way no know which values are
    removed. So use it carefuly. On the contrary, zadd, zrem and zincrby are
    optimized to only index/deindex updated values
    """

    proxy_getter = "zmembers"
    proxy_setter = "zadd"

    available_getters = ('zcard', 'zcount', 'zrange', 'zrangebyscore',
                         'zrank', 'zrevrange', 'zrevrangebyscore',
                         'zrevrank', 'zscore', 'zscan', 'sort')
    available_modifiers = ('delete', 'zadd', 'zincrby', 'zrem',
                           'zremrangebyrank', 'zremrangebyscore', )

    _call_zrem = MultiValuesField._rem
    _call_zremrangebyscore = _call_zremrangebyrank = RedisField._reset

    scannable = True
    _call_zscan = MultiValuesField._scan
    if redispy_version >= (2, 10, 0):
        _call_zscan_iter = MultiValuesField._scan
        available_getters = available_getters + ('zscan_iter', )

    def zmembers(self):
        """
        Used as a proxy_getter to get all values stored in the field.
        """
        return self.zrange(0, -1)

    def _call_zadd(self, command, *args, **kwargs):
        """
        We do the same computation of the zadd method of StrictRedis to keep keys
        to index them (instead of indexing the whole set)
        Members (value/score) can be passed:
            - in *args, with score followed by the value, 0+ times (to respect
              the redis order)
            - in **kwargs, with value as key and score as value
        Example: zadd(1.1, 'my-key', 2.2, 'name1', 'name2', name3=3.3, name4=4.4)
        """
        if self.indexable:
            keys = []
            if args:
                if len(args) % 2 != 0:
                    raise RedisError("ZADD requires an equal number of "
                                     "values and scores")
                keys.extend(args[1::2])
            keys.extend(kwargs)  # add kwargs keys (values to index)
            self.index(keys)
        return self._traverse_command(command, *args, **kwargs)

    def _call_zincrby(self, command, value, *args, **kwargs):
        """
        This command update a score of a given value. But it can be a new value
        of the sorted set, so we index it.
        """
        if self.indexable:
            self.index([value])
        return self._traverse_command(command, value, *args, **kwargs)

    @staticmethod
    def coerce_zadd_args(*args, **kwargs):
        """
        Take arguments attended by a zadd call, named or not, and return a flat list
        that can be used.
        A callback can be called with all "values" (as *args) if defined as the
        `values_callback` named argument. Real values will then be the result of
        this callback.
        """
        values_callback = kwargs.pop('values_callback', None)

        pieces = []
        if args:
            if len(args) % 2 != 0:
                raise RedisError("ZADD requires an equal number of "
                                 "values and scores")
            pieces.extend(args)

        for pair in iteritems(kwargs):
            pieces.append(pair[1])
            pieces.append(pair[0])

        values = pieces[1::2]
        if values_callback:
            values = values_callback(*values)

        scores = pieces[0::2]

        pieces = []
        for z in zip(scores, values):
            pieces.extend(z)

        return pieces


class SetField(MultiValuesField):
    """
    A field with values stored in a redis set.
    If the indexable argument is set to True on the constructor, all stored
    values will be indexed.
    sadd, srem and spop commands are optimized to index/deindex only needed values
    """

    proxy_getter = "smembers"
    proxy_setter = "sadd"

    available_getters = ('scard', 'sismember', 'smembers', 'srandmember', 'sscan', 'sort', )
    available_modifiers = ('delete', 'sadd', 'srem', 'spop', )

    _call_sadd = MultiValuesField._add
    _call_srem = MultiValuesField._rem
    _call_spop = MultiValuesField._pop

    scannable = True
    _call_sscan = MultiValuesField._scan
    if redispy_version >= (2, 10, 0):
        _call_sscan_iter = MultiValuesField._scan
        available_getters = available_getters + ('sscan_iter', )


class ListField(MultiValuesField):
    """
    A field with values stored in a list.
    If the indexable argument is set to True on the constructor, all stored
    values will be indexed (one entry in the index for each value even if the
    value is stored many times in the list). But when using ltrim, all content
    will be deindexed and then reindexed as we have no way no know which values
    are removed. So use it carefuly. On the contrary, linsert, *pop, *push*,
    lset are optimized to only index/deindex updated values. lrem is optimized
    only if the "count" attribute is set to 0
    """

    proxy_getter = "lmembers"
    proxy_setter = "rpush"

    available_getters = ('lindex', 'llen', 'lrange', )
    available_modifiers = ('delete', 'linsert', 'lpop', 'lpush', 'lpushx',
                           'lrem', 'rpop', 'rpush', 'rpushx', 'lset',
                           'ltrim', 'sort', )

    _call_lpop = _call_rpop = MultiValuesField._pop
    _call_lpush = _call_rpush = MultiValuesField._add
    _call_ltrim = RedisField._reset

    def lmembers(self):
        """
        Used as a proxy_getter to get all values stored in the field.
        """
        return self.lrange(0, -1)

    def _pushx(self, command, *args, **kwargs):
        """
        Helper for lpushx and rpushx, that only index the new values if the list
        existed when the command was called
        """
        result = self._traverse_command(command, *args, **kwargs)
        if self.indexable and result:
            self.index(args)
        return result
    _call_lpushx = _pushx
    _call_rpushx = _pushx

    def _call_lrem(self, command, count, value, *args, **kwargs):
        """
        If count is 0, we remove all elements equal to value, so we know we have
        nothing to index, and this value to deindex. In other case, we don't
        know how much elements will remain in the list, so we have to do a full
        deindex/reindex. So do it carefuly.
        """
        if not count:
            if self.indexable:
                self.deindex([value])
            return self._traverse_command(command, count, value, *args, **kwargs)
        else:
            return self._reset(command, count, value, *args, **kwargs)

    def _call_lset(self, command, index, value, *args, **kwargs):
        """
        Before setting the new value, get the previous one to deindex it. Then
        call the command and index the new value, if exists
        """
        if self.indexable:
            old_value = self.lindex(index)
            self.deindex([old_value])
        result = self._traverse_command(command, index, value, *args, **kwargs)
        if self.indexable:
            self.index([value])
        return result

    def _call_linsert(self, command, where, refvalue, value):
        result = self._traverse_command(command, where, refvalue, value)
        if self.indexable and result != -1:
            self.index([value])
        return result


class HashField(MultiValuesField):

    _unique_supported = False
    _field_parts = 2

    proxy_getter = "hgetall"
    proxy_setter = "hmset"

    available_getters = ('hget', 'hgetall', 'hmget', 'hkeys', 'hvals',
                         'hlen', 'hscan', )
    available_modifiers = ('delete', 'hdel', 'hmset', 'hsetnx', 'hset',
                           'hincrby', 'hincrbyfloat', )

    scannable = True
    _call_hscan = MultiValuesField._scan
    if redispy_version >= (2, 10, 0):
        _call_hscan_iter = MultiValuesField._scan
        available_getters = available_getters + ('hscan_iter', )

    def _call_hmset(self, command, *args, **kwargs):
        if self.indexable:
            current = self.proxy_get()
            _to_deindex = dict((k, current[k]) for k in iterkeys(kwargs) if k in current)
            self.deindex(_to_deindex)
            self.index(kwargs)
        return self._traverse_command(command, kwargs)

    def _call_hset(self, command, key, value):
        if self.indexable:
            current = self.hget(key)
            if current != value:
                if current is not None:
                    self.deindex({key: current})
                if value is not None:
                    self.index({key: value})
        return self._traverse_command(command, key, value)

    def _call_hincrby(self, command, key, amount):
        if self.indexable:
            current = self.hget(key)
            if current is not None:
                self.deindex({key: current})
        result = self._traverse_command(command, key, amount)
        if self.indexable:
            self.index({key: result})
        return result
    _call_hincrbyfloat = _call_hincrby

    def _call_hdel(self, command, *args):
        if self.indexable:
            current = self.proxy_get()
            self.deindex(dict((k, current[k]) for k in args if k in current))
        return self._traverse_command(command, *args)

    def _call_hsetnx(self, command, key, value):
        result = self._traverse_command(command, key, value)
        if self.indexable and result:
            # hsetnx returns 1 if key has been set
            self.index({key: value})
        return result

    def _call_hmget(self, command, *args):
        # redispy needs a list, not args
        return self._traverse_command(command, args)

    def index(self, values=None, only_index=None):
        """
        Deal with dicts and field names.
        """
        assert self.indexable, "Field not indexable"
        assert not only_index or self.has_index(only_index), "Invalid index"
        if only_index:
            only_index = only_index if isclass(only_index) else only_index.__class__

        if values is None:
            values = self.proxy_get()

        for field_name, value in iteritems(values):
            if value is not None:
                needs_to_check_uniqueness = bool(self.unique)

                for index in self._indexes:
                    if only_index and not isinstance(index, only_index):
                        continue

                    index.add(field_name, value, check_uniqueness=needs_to_check_uniqueness and index.handle_uniqueness)

                    if needs_to_check_uniqueness and index.handle_uniqueness:
                        # uniqueness check is done for this value
                        needs_to_check_uniqueness = False

    def deindex(self, values=None, only_index=None):
        """
        Deal with dicts and field names.
        """
        assert self.indexable, "Field not indexable"
        assert not only_index or self.has_index(only_index), "Invalid index"
        if only_index:
            only_index = only_index if isclass(only_index) else only_index.__class__

        if values is None:
            values = self.proxy_get()
        for field_name, value in iteritems(values):
            if value is not None:
                for index in self._indexes:
                    if only_index and not isinstance(index, only_index):
                        continue
                    index.remove(field_name, value)

    def hexists(self, key):
        """
        Call the hexists command to check if the redis hash key exists for the
        current field
        """
        try:
            hashkey = self.key
        except DoesNotExist:
            """
            If the object doesn't exists anymore, its PK is deleted, so the
            "self.key" call will raise a DoesNotExist exception. We catch it
            to return False, as the field doesn't exists too.
            """
            return False
        else:
            return self.connection.hexists(hashkey, key)
    exists = hexists


class InstanceHashField(SingleValueField):
    """Field stored in the parent object hash."""

    proxy_getter = "hget"
    proxy_setter = "hset"

    available_getters = ('hget', )
    available_modifiers = ('hdel', 'hset', 'hsetnx', 'hincrby',
                           'hincrbyfloat', )

    _call_hset = SingleValueField._call_set
    _call_hdel = RedisField._del

    @property
    def key(self):
        return self._instance.key

    @property
    def sort_wildcard(self):
        return "%s->%s" % (self._model.sort_wildcard(), self.name)

    def _traverse_command(self, name, *args, **kwargs):
        """Add key AND the hash field to the args, and call the Redis command."""
        args = list(args)
        args.insert(0, self.name)
        return super(InstanceHashField, self)._traverse_command(name, *args, **kwargs)

    def delete(self):
        """
        Delete the field from redis, only the hash entry
        """
        return self._call_command('hdel')

    def hexists(self):
        """
        Call the hexists command to check if the redis hash key exists for the
        current field
        """
        try:
            key = self.key
        except DoesNotExist:
            """
            If the object doesn't exists anymore, its PK is deleted, so the
            "self.key" call will raise a DoesNotExist exception. We catch it
            to return False, as the field doesn't exists too.
            """
            return False
        else:
            return self.connection.hexists(key, self.name)
    exists = hexists


class PKField(SingleValueField):
    """
    This type of field is used as a primary key.
    There must be one, and only one instance of this field (or a subclass) on a
    model.
    If no PKField is defined on a model, an AutoPKField is automatically added.
    A PKField has no auto-increment, a pk must be passed to constructor.
    """

    # Use only a simple getter and setter. We take all control on the setter.
    proxy_getter = "get"
    proxy_setter = "set"

    available_getters = ('get',)
    available_modifiers = ('set',)

    name = 'pk'  # Default name ok the pk, can be changed by declaring a new PKField
    indexable = False  # Not an `indexable` field...
    unique = True  # ... but `unique` can be usefull in loops
    _auto_increment = False  # False for PKField, True for AutoPKField
    _auto_added = False  # True only if automatically added by limpyd
    _set = False  # True when set for the first (and unique) time

    _copy_conf = copy(RedisField._copy_conf)
    _copy_conf['attrs'] += ['_auto_increment', '_auto_added']

    def normalize(self, value):
        """
        Simple method to always have the same kind of value
        It can be overriden by converting to int
        """
        return str(value)

    def _validate(self, value):
        """
        Validate that a given new pk to set is always set, and return it.
        The returned value should be normalized, and will be used without check.
        """
        if value is None:
            raise ValueError('The pk for %s is not "auto-increment", you must fill it' %
                            self._model.__name__)
        value = self.normalize(value)

        # Check that this pk does not already exist
        if self.exists(value):
            raise UniquenessError('PKField %s already exists for model %s)' %
                                  (value, self._instance.__class__))

        return value

    @property
    def collection_key(self):
        """
        Property that return the name of the key in Redis where are stored
        all the exinsting pk for the model hosting this PKField
        """
        return '%s:collection' % self._model._name

    def exists(self, value=None):
        """
        Return True if the given pk value exists for the given class.
        If no value is given, we use the value of the current field, which
        is the value of the "_pk" attribute of its instance.
        """
        try:
            if not value:
                value = self.get()
        except (AttributeError, DoesNotExist):
            # If the instance is deleted, the _pk attribute doesn't exist
            # anymore. So we catch the AttributeError to return False (this pk
            # field doesn't exist anymore) in this specific case
            return False
        else:
            return self.connection.sismember(self.collection_key, value)

    def collection(self):
        """
        Return all available primary keys for the given class
        """
        return self.connection.smembers(self.collection_key)

    def set(self, value):
        """
        Override the default setter to check uniqueness, deny updating, and add
        the new pk to the model's collection.
        The value is not saved as a field in redis, because we don't need this.
        On an instance, we have the _pk attribute with the pk value, and when
        we ask for a collection, we get somes pks which can be used to instanciate
        new objects (holding the pk value in _pk)
        """
        # Deny updating of an already set pk
        if self._set:
            raise ValueError('A primary key cannot be updated')

        # Validate and return the value to be used as a pk
        value = self._validate(value)

        # Tell the model the pk is now set
        self._instance._set_pk(value)
        self._set = True

        # We have a new pk, so add it to the collection
        log.debug("Adding %s in %s collection" % (value, self._model._name))
        self.connection.sadd(self.collection_key, value)

        # Finally return 1 as we did a real redis call to the set command
        return 1

    def get(self):
        """
        We do not call the default getter as we have the value cached in the
        instance in its _pk attribute
        """
        if not hasattr(self, '_instance'):
            raise ImplementationError("Impossible to get the PK of an unbound field")
        if not hasattr(self._instance, '_pk'):
            raise DoesNotExist("The current object doesn't exists anymore")

        if not self._instance._pk:
            self.set(value=None)

        return self.normalize(self._instance._pk)

    def delete(self):
        raise ImplementationError('PKField cannot be deleted directly.')


class AutoPKField(PKField):
    """
    A subclass of PKField that implement auto-increment. Models with an
    AutoPKField cannot pass pk to constructors, they are always set by
    incrementing the last pk used
    """
    _auto_increment = True

    def _validate(self, value):
        """
        Validate that a given new pk to set is always set to None, then return
        a new pk
        """
        if value is not None:
            raise ValueError('The pk for %s is "auto-increment", you must not fill it' %
                            self._model.__name__)
        key = self._instance.make_key(self._model._name, 'max_pk')
        return self.normalize(self.connection.incr(key))


class FieldLock(Lock):
    """
    This subclass of the Lock object is used to add a lock on the field. It will
    be used on write operations to block writes for other instances on this
    field, during all operations needed to do a deindex+write+index.
    Only one lock is done on a specific field for a specific model. If during
    lock, another one is asked in the same thread, we assume that it's a
    operation that must be done during the main lock and we don't wait for
    release.
    """

    def __init__(self, field, timeout=5, sleep=0.1):
        """
        Save the field and create a real lock,, using the correct connection
        and a computed lock key based on the names of the field and its model.
        """
        self.field = field
        self.sub_lock_mode = False
        super(FieldLock, self).__init__(
            redis=field._model.get_connection(),
            name=make_key(field._model._name, 'lock-for-update', field.name),
            timeout=timeout,
            sleep=sleep,
        )

    def _get_already_locked_by_model(self):
        """
        A lock is self_locked if already set for the current field+model on the current
        thread.
        """
        return self.field._model._is_field_locked(self.field)

    def _set_already_locked_by_model(self, value):
        if value:
            self.field._model._mark_field_as_locked(self.field)
        else:
            self.field._model._unmark_field_as_locked(self.field)

    already_locked_by_model = property(_get_already_locked_by_model, _set_already_locked_by_model)

    def acquire(self, *args, **kwargs):
        """
        Really acquire the lock only if it's not a sub-lock. Then save the
        sub-lock status.
        """
        if not self.field.lockable:
            return
        if self.already_locked_by_model:
            self.sub_lock_mode = True
            return
        self.already_locked_by_model = True
        super(FieldLock, self).acquire(*args, **kwargs)

    def release(self, *args, **kwargs):
        """
        Really release the lock only if it's not a sub-lock. Then save the
        sub-lock status and mark the model as unlocked.
        """
        if not self.field.lockable:
            return
        if self.sub_lock_mode:
            return
        super(FieldLock, self).release(*args, **kwargs)
        self.already_locked_by_model = self.sub_lock_mode = False

    def __exit__(self, *args, **kwargs):
        """
        Mark the model as unlocked.
        """
        super(FieldLock, self).__exit__(*args, **kwargs)
        if not self.field.lockable:
            return
        if not self.sub_lock_mode:
            self.already_locked_by_model = False



"""Limpyd provides an easy way to store objects in Redis, without losing the
power and the control of the Redis API, in a limpid way, with just as
abstraction as needed."""

import pkg_resources
from future.builtins import str
from os import path
from setuptools.config import read_configuration


def _extract_version(package_name):
    try:
        # if package is installed
        version = pkg_resources.get_distribution(package_name).version
    except pkg_resources.DistributionNotFound:
        # if not installed, so we must be in source, with ``setup.cfg`` available
	    _conf = read_configuration(path.join(
	        path.dirname(__file__), '..', 'setup.cfg')
	    )
	    version = _conf['metadata']['version']

    return version


EXACT_VERSION = _extract_version('redis_limpyd')
VERSION = tuple(int(part) for part in EXACT_VERSION.split('.') if str(part).isnumeric())



from __future__ import unicode_literals

__all__ = [
    "UniquenessError",
    "ImplementationError",
    "DoesNotExist",
]


class LimpydException(Exception):
    pass


class UniquenessError(LimpydException):
    pass


class DoesNotExist(LimpydException):
    pass


class ImplementationError(LimpydException):
    pass



# -*- coding:utf-8 -*-
from __future__ import unicode_literals
from future.builtins import map
from future.builtins import str
from future.builtins import object
from future.utils import with_metaclass

import re
from copy import copy

from limpyd import model, fields
from limpyd.exceptions import *
from limpyd.contrib.collection import ExtendedCollectionManager

# used to validate a related_name
re_identifier = re.compile(r"\W")


class RelatedCollection(object):
    """
    When a related field is added, a related collection is created on the
    related model on the other side.
    It's simply a shortcut to a collection with a predefined filter.

    Exemple with these two related classes :

        class Person(RelatedModel):
            name = FKStringField()
            group = FKStringField('Group', related_name='members')

        class Group(RelatedField):
            name = FKStringField()

    Defining these objets :

        group = Group(name='a group')
        person = Person(name='a person', group=group)

    You can access members of the group via the main way:

        members = Person.collection(group=group._pk)

    Or with the related collection:

        members = group.members()

    Note that you can pass filters the same way you can pass them to a collection:
        members = group.members(a_filter=a_value, another_filter=another_value)

    The result is a real collection, with lazy loading, sorting...
    """

    def __init__(self, instance, related_field):
        """
        Create the RelatedCollection on the 'instance', related to the field
        'related_field'
        """
        self.instance = instance
        self.related_field = related_field

    def __call__(self, **filters):
        """
        Return a collection on the related model, given the current instance as
        a filter for the related field.
        """
        if not filters:
            filters = {}
        filters[self.related_field.name] = self.instance._pk
        return self.related_field._model.collection(**filters)

    def remove_instance(self):
        """
        Remove the instance from the related fields (delete the field if it's
        a simple one, or remove the instance from the field if it's a set/list/
        sorted_set)
        """
        with fields.FieldLock(self.related_field):
            related_pks = self()
            for pk in related_pks:

                # get the real related field
                related_instance = self.related_field._model(pk)
                related_field = getattr(related_instance, self.related_field.name)

                # check if we have a dedicated remove method
                remover = getattr(related_field, '_related_remover', None)

                # then remove the instance from the related field
                if remover is not None:
                    # if we have a remover method, it wants the instance as argument
                    # (the related field may be a set/list/sorted_set)
                    getattr(related_field, remover)(self.instance._pk)
                else:
                    # no remover method, simple delete the field
                    related_field.delete()


class RelatedModel(model.RedisModel):
    """
    This subclass of RedisModel handles creation of related collections, and
    propagates to them the deletion of the instance. So it's needed for models
    with related fields to subclass this RelatedModel instead of RedisModel.
    """

    abstract = True
    collection_manager = ExtendedCollectionManager

    def __init__(self, *args, **kwargs):
        """
        Create the instance then add all related collections (link between this
        instance and related fields on other models)
        """
        super(RelatedModel, self).__init__(*args, **kwargs)

        # create the related collections
        self.related_collections = []
        relations = getattr(self.database, '_relations', {}).get(self._name.lower(), [])
        for relation in relations:
            # get the related field
            model_name, field_name, _ = relation
            related_model = self.database._models[model_name]
            related_field = related_model.get_field(field_name)

            # add the collection
            collection = related_field.related_collection_class(self, related_field)
            setattr(self, related_field.related_name, collection)
            self.related_collections.append(related_field.related_name)

    def delete(self):
        """
        When the instance is deleted, we propagate the deletion to the related
        collections, which will remove it from the related fields.
        """
        for related_collection_name in self.related_collections:
            related_collection = getattr(self, related_collection_name)
            related_collection.remove_instance()
        return super(RelatedModel, self).delete()

    @classmethod
    def use_database(cls, database):
        """
        Move model and its submodels to the new database, as the original
        use_database method. And manage relations too (done here instead of
        the database because the database class is not aware of relations)
        """
        original_database = getattr(cls, 'database', None)
        impacted_models = super(RelatedModel, cls).use_database(database)

        # no relation on the current database, so nothing nore to transfer
        if not original_database or not getattr(original_database, '_relations', {}):
            return impacted_models

        # prepare relations to remove impacted ones
        # relations in original_database._relations have this format
        # related_model => (original_model, original_field, related_name)
        # Here we want all relations with an original_model in our list of
        # models impacted by the change of database (impacted_models), to move
        # these relation on the new database
        reverse_relations = {}
        for related_model_name, relations in original_database._relations.items():
            for relation in relations:
                reverse_relations.setdefault(relation[0], []).append((related_model_name, relation))

        # create an dict to store relations in the new database
        if not hasattr(database, '_relations'):
            database._relations = {}

        # move relation for all impacted models
        for _model in impacted_models:
            if _model.abstract:
                continue
            for related_model_name, relation in reverse_relations[_model._name]:
                # if the related model name is already used as a relation, check
                # if it's not already used with the related_name of the relation
                if related_model_name in database._relations:
                    field = _model.get_field(relation[1])
                    field._assert_relation_does_not_exists()
                # move the relation from the original database to the new
                original_database._relations[related_model_name].remove(relation)
                database._relations.setdefault(related_model_name, []).append(relation)

        return impacted_models


class RelatedFieldMetaclass(fields.MetaRedisProxy):
    """
    Metaclass for RelatedFieldMixin that will create methods for all commands
    defined in "_commands_with_single_value_from_python" and
    "_commands_with_many_values_from_python", to override the default
    behaviour: convert values given as object to their pk, before calling the
    super method
    """

    def __new__(mcs, name, base, dct):
        it = super(RelatedFieldMetaclass, mcs).__new__(mcs, name, base, dct)
        for command_name in it._commands_with_single_value_from_python:
            setattr(it, command_name, it._make_command_method(command_name, many=False))
        for command_name in it._commands_with_many_values_from_python:
            setattr(it, command_name, it._make_command_method(command_name, many=True))
        return it


class RelatedFieldMixin(with_metaclass(RelatedFieldMetaclass)):
    """
    Base mixin for all fields holding related instances.
    This mixin provides:
    - force indexable to True
    - a "from_python" method that can translate a RedisModel instance, or a
      FK field in its pk (useful to pass object instead of "object._pk" when
      adding/removing a FK)
      All commands that may receive objects as arguments must call this
      "from_python" method (or "from_python_many" if many values to convert).
      To do this automatically, simply add command names that accept only one
      value in "_commands_with_single_value_from_python" and ones that accept
      many values (without any other arguments) in
      "_commands_with_many_values_from_python" (see RelatedFieldMetaclass)
    - management of related parameters: "to" and "related_name"
    """

    _copy_conf = copy(fields.RedisField._copy_conf)
    _copy_conf['kwargs'] += [('to', 'related_to'), 'related_name']

    _commands_with_single_value_from_python = []
    _commands_with_many_values_from_python = []

    related_collection_class = RelatedCollection

    def __init__(self, to, *args, **kwargs):
        """
        Force the field to be indexable and save related arguments.
        """
        kwargs['indexable'] = True
        super(RelatedFieldMixin, self).__init__(*args, **kwargs)

        self.related_to = to
        self.related_name = kwargs.pop('related_name', None)

    def _attach_to_model(self, model):
        """
        When we have a model, save the relation in the database, to later create
        RelatedCollection objects in the related model
        """
        super(RelatedFieldMixin, self)._attach_to_model(model)

        if model.abstract:
            # do not manage the relation if it's an abstract model
            return

        # now, check related_name and save the relation in the database

        # get related parameters to identify the relation
        self.related_name = self._get_related_name()
        self.related_to = self._get_related_model_name()

        # create entry for the model in the _relations list of the database
        if not hasattr(self.database, '_relations'):
            self.database._relations = {}
        self.database._relations.setdefault(self.related_to, [])

        # check unicity of related name for related model
        self._assert_relation_does_not_exists()

        # the relation didn't exists, we can save it
        relation = (self._model._name, self.name, self.related_name)
        self.database._relations[self.related_to].append(relation)

    def _assert_relation_does_not_exists(self):
        """
        Check if a relation with the current related_name doesn't already exists
        for the related model
        """
        relations = self.database._relations[self.related_to]
        existing = [r for r in relations if r[2] == self.related_name]
        if existing:
            error = ("The related name defined for the field '%s.%s', named '%s', already exists "
                     "on the model '%s' (tied to the field '%s.%s')")
            raise ImplementationError(error % (self._model.__name__, self.name, self.related_name,
                                      self.related_to, existing[0][1], existing[0][0]))

    def _get_related_model_name(self):
        """
        Return the name of the related model, as used to store all models in the
        database object, in the following format: "%(namespace)s:%(class_name)"
        The result is computed from the "to" argument of the RelatedField
        constructor, stored in the "related_to" attribute, based on theses rules :
        - if a "RelatedModel" subclass, get its namespace:name
        - if a string :
          - if "self", get from the current model (relation on self)
          - if a RelatedModel class name, keep it, and if no namesace, use the
            namespace of the current model
        """
        if isinstance(self.related_to, type) and issubclass(self.related_to, RelatedModel):
            model_name = self.related_to._name

        elif isinstance(self.related_to, str):
            if self.related_to == 'self':
                model_name = self._model._name
            elif ':' not in self.related_to:
                model_name = ':'.join((self._model.namespace, self.related_to))
            else:
                model_name = self.related_to

        else:
            raise ImplementationError("The `to` argument to a related field "
                                      "must be a RelatedModel as a class or as "
                                      "a string (with or without namespace). "
                                      "Or simply 'self'.")

        return model_name.lower()

    def _get_related_name(self):
        """
        Return the related name to use to access this related field.
        If the related_name argument is not defined in its declaration,
        a new one will be computed following this format: '%s_set' with %s the
        name of the model owning the related.field.
        If the related_name argument exists, it can be the exact name to use (
        be careful to use a valid python attribute name), or a string where you
        can set placeholder for the namespace and the model of the current
        model. It's useful if the current model is abstract with many subclasses

        Exemples:
            class Base(RelatedModel):
                abstract = True
                namespace = 'project'
                a_field = FKStringField('Other', related_name='%(namespace)s_%(model)s_related')
                    # => related names accessible from Other are
                    #    "project_childa_related" and "project_childb_related"

            class ChildA(Base):
                pass

            class ChildB(Base):
                pass

            class Other(RelatedModel):
                namespace = 'project'
                field_a = FKStringField(ChildA)
                    # => related name accessible from ChildA and ChildB will be "other_related"
                field_b = FKStringField(ChildB, related_name='some_others')
                    # => related name accessible from ChildA and ChildB will be "some_others"
        """
        related_name = self.related_name or '%(model)s_set'
        related_name = related_name % {
            'namespace': re_identifier.sub('_', self._model.namespace),
            'model': self._model.__name__
        }
        if re_identifier.sub('', related_name) != related_name:
            raise ImplementationError('The related_name "%s" is not a valid '
                                      'python identifier' % related_name)

        return related_name.lower()

    def from_python(self, value):
        """
        Provide the ability to pass a RedisModel instances or a FK field as
        value instead of passing the PK. The value will then be translated in
        the real PK.
        """
        if isinstance(value, model.RedisModel):
            value = value._pk
        elif isinstance(value, SimpleValueRelatedFieldMixin):
            value = value.proxy_get()
        return value

    def from_python_many(self, *values):
        """
        Apply the from_python to each values and return the final list
        """
        return list(map(self.from_python, values))

    @classmethod
    def _make_command_method(cls, command_name, many=False):
        """
        Return a function which will convert objects to their pk, then call the
        super method for the given name.
        The "many" attribute indicates that the command accept one or many
        values as arguments (in *args)
        """
        def func(self, *args, **kwargs):
            if many:
                args = self.from_python_many(*args)
            else:
                if 'value' in kwargs:
                    kwargs['value'] = self.from_python(kwargs['value'])
                else:
                    args = list(args)
                    args[0] = self.from_python(args[0])

            # call the super method, with real pk
            sup_method = getattr(super(cls, self), command_name)
            return sup_method(*args, **kwargs)
        return func


class SimpleValueRelatedFieldMixin(RelatedFieldMixin):
    """
    Mixin for all related fields storing one unique value.
    Add a instance method to get the related object.
    Example:

        class Person(RelatedModel):
            name = PKField()

        class Group(RelatedModel):
            name = PKField()
            owner = FKStringField(Person, related_name='owned_groups')

        person = Person(name='person')
        group = Group(name='group', owner=person)

        # returns the owner's pk
        group.owner.get()

        # return the full person instance
        group.owner.instance()

    """
    def instance(self, skip_exist_test=False):
        """
        Returns the instance of the related object linked by the field.
        """
        model = self.database._models[self.related_to]
        meth = model.lazy_connect if skip_exist_test else model
        return meth(self.proxy_get())


class FKStringField(SimpleValueRelatedFieldMixin, fields.StringField):
    """ Related field based on a StringField, acting as a Foreign Key """
    _commands_with_single_value_from_python = ['set', 'setnx', 'getset', ]


class FKInstanceHashField(SimpleValueRelatedFieldMixin, fields.InstanceHashField):
    """ Related field based on a InstanceHashField, acting as a Foreign Key """
    _commands_with_single_value_from_python = ['hset', 'hsetnx', ]


class MultiValuesRelatedFieldMixin(RelatedFieldMixin):
    """
    Mixin for all related fields based on MultiValuesField.
    Add a __call__ method creating a collection to use the field the same way
    we can use the RelatedCollection on the other side.
    Example showing both sides:

        class Person(RelatedModel):
            name = PKField()

        class Group(RelatedModel):
            name = PKField()
            members = M2MSetField(Person, related_name='membership')

        person1 = Person(name='person1')
        person2 = Person(name='person2')
        group1 = Group(name='group1')
        group2 = Group(name='group2')

        group1.members.sadd(person1, person2)
        group2.members.sadd(person1, person2)

        # A RelatedColleciton that return a collection with set(['person1', 'person2'])
        person1.membership()
        # A M2MSetField call, that return a collection with set(['group1', 'group2'])
        group1.members()

    """

    def __call__(self, **filters):
        """
        When calling (via `()` or via `.collection()`) a MultiValuesRelatedField,
        we return a collection, filtered with given arguments, the result beeing
        "intersected" with the members of the current field.
        """
        model = self.database._models[self.related_to]
        collection = model.collection_manager(model)
        return collection(**filters).intersect(self)

    # calling obj.field.collection() is the same as calling obj.field()
    collection = __call__


class M2MSetField(MultiValuesRelatedFieldMixin, fields.SetField):
    """ Related field based on a SetField, acting as a M2M """
    _commands_with_single_value_from_python = ['sismember', ]
    _commands_with_many_values_from_python = ['sadd', 'srem', ]
    _related_remover = 'srem'


class M2MListField(MultiValuesRelatedFieldMixin, fields.ListField):
    """ Related field based on a ListField, acting as a sorted M2M """
    _commands_with_single_value_from_python = ['lpushx', 'rpushx', ]
    _commands_with_many_values_from_python = ['lpush', 'rpush', ]
    _related_remover = 'lrem'

    def linsert(self, where, refvalue, value):
        value = self.from_python(value)
        return super(M2MListField, self).linsert(where, refvalue, value)

    def lrem(self, count, value):
        value = self.from_python(value)
        return super(M2MListField, self).lrem(count, value)

    def lset(self, index, value):
        value = self.from_python(value)
        return super(M2MListField, self).lset(index, value)


class M2MSortedSetField(MultiValuesRelatedFieldMixin, fields.SortedSetField):
    """ Related field based on a SortesSetField, acting as a M2M with scores """
    _commands_with_single_value_from_python = ['zscore', 'zrank', 'zrevrank']
    _commands_with_many_values_from_python = ['zrem']
    _related_remover = 'zrem'

    def zadd(self, *args, **kwargs):
        """
        Parse args and kwargs to check values to pass them through the
        from_python method.
        We pass the parsed args/kwargs as args in the super call, to avoid
        doing the same calculation on kwargs one more time.
        """
        if 'values_callback' not in kwargs:
            kwargs['values_callback'] = self.from_python_many
        pieces = fields.SortedSetField.coerce_zadd_args(*args, **kwargs)
        return super(M2MSortedSetField, self).zadd(*pieces)

    def zincrby(self, value, amount=1):
        value = self.from_python(value)
        return super(M2MSortedSetField, self).zincrby(value, amount)



# -*- coding:utf-8 -*-
from __future__ import unicode_literals
from future.builtins import str
from future.builtins import object

import redis

from limpyd.exceptions import *
from limpyd.indexes import EqualIndex

from logging import getLogger
log = getLogger(__name__)


DEFAULT_CONNECTION_SETTINGS = dict(
    host="localhost",
    port=6379,
    db=0
)


class RedisDatabase(object):
    """
    A RedisDatabase regroups some models and handles the connection to Redis for
    them.
    Each model must have a database entry, but many (or all) can share the same
    RedisDatabase object (so each of these models will be stored on the same
    Redis server+database)
    In a database, two models with the same namespace (empty by default) cannot
    have the same name (defined by the class name)
    """
    _connections = {}  # class level cache

    default_indexes = [EqualIndex]

    def __init__(self, **connection_settings):
        self._connection = None  # Instance level cache
        self.reset(**(connection_settings or DEFAULT_CONNECTION_SETTINGS))
        # _models keep an entry for each defined model on this database
        self._models = dict()
        super(RedisDatabase, self).__init__()

    @classmethod
    def get_default_indexes(cls):
        if cls.default_indexes is not None:
            return cls.default_indexes
        return []

    def connect(self, **settings):
        """
        Connect to redis and cache the new connection
        """
        # compute a unique key for this settings, for caching. Work on the whole
        # dict without directly using known keys to allow the use of unix socket
        # connection or any other (future ?) way to connect to redis
        if not settings:
            settings = self.connection_settings
        connection_key = ':'.join([str(settings[k]) for k in sorted(settings)])
        if connection_key not in self._connections:
            self._connections[connection_key] = redis.StrictRedis(
                                            decode_responses=True, **settings)
        return self._connections[connection_key]

    def reset(self, **connection_settings):
        """
        Set the new connection settings to be used and reset the connection
        cache so the next redis call will use these settings.
        """
        self.connection_settings = connection_settings
        self._connection = None

    def _add_model(self, model):
        """
        Save this model as one existing on this database, to deny many models
        with same namespace and name.
        If the model already exists, check if it is the same. It can happen if the
        module is imported twice in different ways.
        If it's a new model or an existing and valid one, return the model in database: the
        one added or the existing one
        """
        name = model._name
        existing = self._models.get(name, None)
        if not existing:
            self._models[name] = model
        elif model.__name__ != existing.__name__ or model._creation_source != existing._creation_source:
            raise ImplementationError(
                'A model with namespace "%s" and name "%s" is already defined '
                'on this database' % (model.namespace, model.__name__))
        return self._models[name]

    def _use_for_model(self, model):
        """
        Update the given model to use the current database. Do it also for all
        of its subclasses if they share the same database. (so it's easy to
        call use_database on an abstract model to use the new database for all
        subclasses)
        """
        original_database = getattr(model, 'database', None)

        def get_models(model):
            """
            Return the model and all its submodels that are on the same database
            """
            model_database = getattr(model, 'database', None)
            if model_database == self:
                return []
            models = [model]
            for submodel in model.__subclasses__():
                if getattr(submodel, 'database', None) == model_database:
                    models += get_models(submodel)
            return models

        # put the model and all its matching submodels on the new database
        models = get_models(model)
        for _model in models:
            if not _model.abstract:
                self._add_model(_model)
                del original_database._models[_model._name]
            _model.database = self

        # return updated models
        return models

    @property
    def connection(self):
        """
        A simple property on the instance that return the connection stored on
        the class
        """
        if self._connection is None:
            self._connection = self.connect()
        return self._connection

    @property
    def redis_version(self):
        """Return the redis version as a tuple"""
        if not hasattr(self, '_redis_version'):
            self._redis_version = tuple(
                map(int, self.connection.info().get('redis_version').split('.')[:3])
            )
        return self._redis_version

    def support_scripting(self):
        """
        Returns True if scripting is available. Checks are done in the client
        library (redis-py) AND the redis server. Result is cached, so done only
        one time.
        """
        if not hasattr(self, '_support_scripting'):
            try:
                self._support_scripting = self.redis_version >= (2, 5) \
                    and hasattr(self.connection, 'register_script')
            except:
                self._support_scripting = False
        return self._support_scripting

    def support_zrangebylex(self):
        """
        Returns True if zrangebylex is available. Checks are done in the client
        library (redis-py) AND the redis server. Result is cached, so done only
        one time.
        """
        if not hasattr(self, '_support_zrangebylex'):
            try:
                self._support_zrangebylex = self.redis_version >= (2, 8, 9) \
                    and hasattr(self.connection, 'zrangebylex')
            except:
                self._support_zrangebylex = False
        return self._support_zrangebylex

    def call_script(self, script_dict, keys=None, args=None):
        """Call a redis script with keys and args

        The first time we call a script, we register it to speed up later calls.
        We expect a dict with a ``lua`` key having the script, and the dict will be
        updated with a ``script_object`` key, with the content returned by the
        the redis-py ``register_script`` command.

        Parameters
        ----------
        script_dict: dict
            A dict with a ``lua`` entry containing the lua code. A new key, ``script_object``
            will be added after that.
        keys: list of str
            List of the keys that will be read/updated by the lua script
        args: list of str
            List of all the args expected by the script.

        Returns
        -------
        Anything that will be returned by the script

        """
        if keys is None:
            keys = []
        if args is None:
            args = []
        if 'script_object' not in script_dict:
            script_dict['script_object'] = self.connection.register_script(script_dict['lua'])
        return script_dict['script_object'](keys=keys, args=args, client=self.connection)

    def scan_keys(self, match=None, count=None):
        """Take a pattern expected by the redis `scan` command and iter on all matching keys

        Parameters
        ----------
        match: str
            The pattern of keys to look for
        count: int, default to None (redis uses 10)
            Hint for redis about the number of expected result

        Yields
        -------
        str
            All keys found by the scan, one by one. A key can be returned multiple times, it's
            related to the way the SCAN command works in redis.

        """
        cursor = 0
        while True:
            cursor, keys = self.connection.scan(cursor, match=match, count=count)
            for key in keys:
                yield key
            if not cursor or cursor == '0':  # string for redis.py < 2.10
                break


class Lock(redis.client.Lock):
    """
    Override the default Lock class to manage the fact that we use ``decode_responses=True``
    but it is not taken into account in python 3 with redis 2.10, as the lock token in
    ``threading.local`` is saved as ``bytes`` but the value retrieved from redis is decoded.
    So the equal check doesn't work.
    So we decode the value in ``threading.local`` to make this check work.
    See https://github.com/andymccurdy/redis-py/issues/694
    """

    def do_release(self, expected_token):

        if isinstance(expected_token, bytes) and \
                self.redis.connection_pool.connection_kwargs.get('decode_responses', False):
            expected_token = expected_token.decode()

        super(Lock, self).do_release(expected_token)



# -*- coding:utf-8 -*-
from __future__ import unicode_literals

import threading

from redis.client import StrictPipeline
from redis.exceptions import WatchError

from limpyd.database import RedisDatabase
from limpyd.fields import RedisField


class PipelineDatabase(RedisDatabase):
    """
    In addition to the functionalities of the default RedisDatabase object, this
    one provide an abstraction to the Pipeline object from
    redis-py, to use its stored connection. This pipeline method returns an
    object, as in redis-py, which provide a watch method attending key names,
    but you can simply provide limpyd fields.
    As for redis-py, a convenience method exists to handle pipeline associated
    to the watch command: transaction.
    """

    def __init__(self, **connection_settings):
        super(PipelineDatabase, self).__init__(**connection_settings)
        self._pipelined_connection = None
        self._direct_connection = None

    def pipeline(self, transaction=True, share_in_threads=False):
        """
        A replacement to the default redis pipeline method which manage saving
        and restoring of the connection.
        ALL calls to redis for the current database will pass via the pipeline.
        So if you don't use watches, all getters will not return any value, but
        results will be available via the pipe.execute() call.
        Please refer to the redis-py documentation.
        To use with "with":
        ###
        # Simple multi/exec
        with database.pipeline() as pipe:
            # do some stuff...
            result = pipe.execute()
        ###
        # Simple pipeline (no transaction, no atomicity)
        with database.pipeline(transaction=False) as pipe:
            # do some stuff...
            result = pipe.execute()
        ###
        # Advanced use with watch
        with database.pipeline() as pipe:
            while 1:
                try:
                    if watches:
                        pipe.watch(watches)
                    # get watched stuff
                    pipe.multi()
                    # do some stuff...
                    return pipe.execute()
                except WatchError:
                    continue
        """
        return _Pipeline(self, transaction=transaction, share_in_threads=share_in_threads)

    def transaction(self, func, *watches, **kwargs):
        """
        Convenience method for executing the callable `func` as a transaction
        while watching all keys specified in `watches`. The 'func' callable
        should expect a single arguement which is a Pipeline object.
        """
        with self.pipeline(True, share_in_threads=kwargs.get('share_in_threads', False)) as pipe:
            while 1:
                try:
                    if watches:
                        pipe.watch(*watches)
                    func(pipe)
                    return pipe.execute()
                except WatchError:
                    continue

    @property
    def _connection(self):
        """
        If we have a pipeline, shared in thread, or in the good thread, return it
        else use the direct connection
        """
        if self._pipelined_connection is not None:
            if self._pipelined_connection.share_in_threads or \
                    threading.current_thread().ident == self._pipelined_connection.current_thread_id:
                return self._pipelined_connection


        return self._direct_connection

    @_connection.setter
    def _connection(self, value):
        """
        If the value is a pipeline, save it as the connection to use for pipelines if to be shared
        in threads or goot thread. Do not remove the direct connection.
        If it is not a pipeline, clear it, and set the direct connection again.
        """
        if isinstance(value, _Pipeline):
            self._pipelined_connection = value
        else:
            self._direct_connection = value
            self._pipelined_connection = None


class _Pipeline(StrictPipeline):
    """
    A subclass of the redis pipeline class used by the database object, which
    save its internal connection and replace it by the pipeline, allowing
    all redis calls to be managed by this pipeline
    """

    def __init__(self, database, transaction=True, share_in_threads=False):
        self._database = database
        self._original_connection = database._connection
        self.share_in_threads = share_in_threads
        if not self.share_in_threads:
            self.current_thread_id = threading.current_thread().ident
        super(_Pipeline, self).__init__(
            connection_pool=database._connection.connection_pool,
            response_callbacks=database._connection.response_callbacks,
            transaction=transaction,
            shard_hint=None)
        database._connection = self

    def watch(self, *names):
        """
        Override the default watch method to allow the user to pass RedisField
        objects as names, which will be translated to their real keys and passed
        to the default watch method
        """
        watches = []
        for watch in names:
            if isinstance(watch, RedisField):
                watch = watch.key
            watches.append(watch)
        return super(_Pipeline, self).watch(*watches)

    def __exit__(self, exc_type, exc_value, traceback):
        self._database._connection = self._original_connection
        super(_Pipeline, self).__exit__(exc_type, exc_value, traceback)



# -*- coding:utf-8 -*-
from __future__ import unicode_literals
from future.builtins import str
from future.builtins import zip
from future.builtins import object

from itertools import islice, chain
from collections import namedtuple
from copy import deepcopy

from limpyd.model import RedisModel
from limpyd.collection import CollectionManager, ParsedFilter
from limpyd.fields import (SetField, ListField, SortedSetField, MultiValuesField,
                           RedisField, SingleValueField)
from limpyd.exceptions import DoesNotExist
from limpyd.contrib.database import PipelineDatabase

SORTED_SCORE = 'sorted_score'
DEFAULT_STORE_TTL = 60


RawFilter = namedtuple('RawFilter', ['name', 'value'])


class ExtendedCollectionManager(CollectionManager):

    _accepted_key_types = {'set', 'zset', 'list'}  # Type of keys indexes are allowed to return

    scripts = {
        'list_to_set': {
            # add all members of the list in a new set
            'lua': """
                redis.call('del', KEYS[2])
                for i, member in ipairs(redis.call('lrange', KEYS[1], 0, -1)) do
                    redis.call('sadd', KEYS[2], member)
                end
                return 1
            """,
        },
    }

    def __init__(self, cls):
        super(ExtendedCollectionManager, self).__init__(cls)

        self._lazy_collection['intersects'] = set()

        self._has_sortedsets = False
        self._sort_by_sortedset = None

        self._store = False
        self.stored_key = False
        self._stored_len = None

        self._values = None  # Will store parameters used to retrieve values

    def _list_to_set(self, list_key, set_key):
        """
        Store all content of the given ListField in a redis set.
        Use scripting if available to avoid retrieving all values locally from
        the list before sending them back to the set
        """
        if self.cls.database.support_scripting():
            self.cls.database.call_script(
                # be sure to use the script dict at the class level
                # to avoid registering it many times
                script_dict=self.__class__.scripts['list_to_set'],
                keys=[list_key, set_key]
            )
        else:
            conn = self.cls.get_connection()
            conn.sadd(set_key, *conn.lrange(list_key, 0, -1))

    @property
    def _collection(self):
        """
        Effectively retrieve data according to lazy_collection.
        If we have a stored collection, without any result, return an empty list
        """
        old_sort_limits_and_len_mode = None if self._sort_limits is None else self._sort_limits.copy(), self._len_mode
        old_sorts = None if self._sort is None else self._sort.copy(),\
                    None if self._sort_by_sortedset is None else self._sort_by_sortedset.copy()
        try:
            if self.stored_key and not self._stored_len:
                if self._len_mode:
                    self._len = 0
                    self._len_mode = False
                self._sort_limits = {}
                return []

            # Manage sort desc added by original `__getitem__` when we sort by score
            if self._sort_by_sortedset and self._sort and self._sort.get('desc'):
                self._sort = None
                self._sort_by_sortedset['desc'] = not self._sort_by_sortedset.get('desc', False)

            return super(ExtendedCollectionManager, self)._collection
        finally:
            self._sort_limits, self._len_mode = old_sort_limits_and_len_mode
            self._sort, self._sort_by_sortedset = old_sorts

    def _prepare_sets(self, sets):
        """
        The original "_prepare_sets" method simple return the list of sets in
        _lazy_collection, know to be all keys of redis sets.
        As the new "intersect" method can accept different types of "set", we
        have to handle them because we must return only keys of redis sets.
        """

        if self.stored_key and not self.stored_key_exists():
            raise DoesNotExist('This collection is based on a previous one, '
                               'stored at a key that does not exist anymore.')

        conn = self.cls.get_connection()

        all_sets = set()
        tmp_keys = set()
        lists = []

        def add_key(key, key_type=None, is_tmp=False):
            if not key_type:
                key_type = conn.type(key)
            if key_type == 'set':
                all_sets.add(key)
            elif key_type == 'zset':
                all_sets.add(key)
                self._has_sortedsets = True
            elif key_type == 'list':
                # if only one list, and no sets, at the end we'll directly use the list
                # else lists will be converted to sets
                lists.append(key)
            elif key_type == 'none':
                # considered as an empty set
                 all_sets.add(key)
            else:
                raise ValueError('Cannot use redis key %s of type %s for filtering' % (
                    key, key_type
                ))
            if is_tmp:
                tmp_keys.add(key)

        for set_ in sets:
            if isinstance(set_, str):
                add_key(set_)
            elif isinstance(set_, ParsedFilter):

                value = set_.value
                # We have a RedisModel and we'll use its pk, or a RedisField
                # (single value) and we'll use its value
                if isinstance(value, RedisModel):
                    value = value.pk.get()
                elif isinstance(value, SingleValueField):
                    value = value.proxy_get()
                elif isinstance(value, RedisField):
                    raise ValueError(u'Invalid filter value for %s: %s' % (set_.index.field.name, value))

                for index_key, key_type, is_tmp in set_.index.get_filtered_keys(
                            set_.suffix,
                            accepted_key_types=self._accepted_key_types,
                            *(set_.extra_field_parts + [value])
                        ):
                    if key_type not in self._accepted_key_types:
                        raise ValueError('The index key returned by the index %s is not valid' % (
                            set_.index.__class__.__name__
                        ))
                    add_key(index_key, key_type, is_tmp)

            elif isinstance(set_, SetField):
                # Use the set key. If we need to intersect, we'll use
                # sunionstore, and if not, store accepts set
                add_key(set_.key, 'set')
            elif isinstance(set_, SortedSetField):
                # Use the sorted set key. If we need to intersect, we'll use
                # zinterstore, and if not, store accepts zset
                add_key(set_.key, 'zset')
            elif isinstance(set_, (ListField, _StoredCollection)):
                add_key(set_.key, 'list')
            elif isinstance(set_, tuple) and len(set_):
                # if we got a list or set, create a redis set to hold its values
                tmp_key = self._unique_key()
                conn.sadd(tmp_key, *set_)
                add_key(tmp_key, 'set', True)
            else:
                raise ValueError('Invalid filter type')

        if lists:
            if not len(all_sets) and len(lists) == 1:
                # only one list, nothing else, we can return the list key
                all_sets = {lists[0]}
            else:
                # we have many sets/lists, we need to convert them to sets
                for list_key in lists:
                    # many sets, convert the list to a simple redis set
                    tmp_key = self._unique_key()
                    self._list_to_set(list_key, tmp_key)
                    add_key(tmp_key, 'set', True)

        return all_sets, tmp_keys

    def filter(self, **filters):
        """
        Add more filters to the collection
        """
        return self._add_filters(**filters)

    def intersect(self, *sets):
        """
        Add a list of sets to the existing list of sets to check. Returns self
        for chaining.
        Each "set" represent a list of pk, the final goal is to return only pks
        matching the intersection of all sets.
        A "set" can be:
        - a string: considered as the name of a redis set, sorted set or list
            (if a list, values will be stored in a temporary set)
        - a list, set or tuple: values will be stored in a temporary set
        - a SetField: we will directly use it's content on redis
        - a ListField or SortedSetField: values will be stored in a temporary
            set (except if we want a sort or values and it's the only "set" to
            use)
        """
        sets_ = set()
        for set_ in sets:
            if isinstance(set_, (list, set)):
                set_ = tuple(set_)
            elif isinstance(set_, MultiValuesField) and not getattr(set_, '_instance', None):
                raise ValueError('%s passed to "intersect" must be bound'
                                 % set_.__class__.__name__)
            elif not isinstance(set_, (tuple, str, MultiValuesField, _StoredCollection)):
                raise ValueError('%s is not a valid type of argument that can '
                                 'be used as a set. Allowed are: string (key '
                                 'of a redis set, sorted set or list), '
                                 'limpyd multi-values field ('
                                 'SetField, ListField or SortedSetField), or '
                                 'real python set, list or tuple' % set_)
            if isinstance(set_, SortedSetField):
                self._has_sortedsets = True
            sets_.add(set_)

        self._lazy_collection['intersects'].update(sets_)
        return self

    def _combine_sets(self, sets, final_set):
        """
        Given a list of set, combine them to create the final set that will be
        used to make the final redis call.
        If we have a least a sorted set, use zinterstore insted of sunionstore
        """
        if self._has_sortedsets:
            self.cls.get_connection().zinterstore(final_set, list(sets))
        else:
            final_set = super(ExtendedCollectionManager, self)._combine_sets(sets, final_set)
        return final_set

    def _final_redis_call(self, final_set, sort_options):
        """
        The final redis call to obtain the values to return from the "final_set"
        with some sort options.
        IIf we have at leaset a sorted set and if we don't have any sort
        options, call zrange on the final set wich is the result of a call to
        zinterstore.
        """
        conn = self.cls.get_connection()

        # we have a sorted set without need to sort, use zrange
        if self._has_sortedsets and sort_options is None:

            return conn.zrange(final_set, 0, -1)

        # we have a stored collection, without other filter, and no need to
        # sort, use lrange
        if self.stored_key and not self._lazy_collection['sets']\
                and len(self._lazy_collection['intersects']) == 1\
                and (sort_options is None or sort_options == {'by': 'nosort'}):

            return conn.lrange(final_set, 0, -1)

        # normal call
        return super(ExtendedCollectionManager, self)._final_redis_call(
                                                        final_set, sort_options)

    def _collection_length(self, final_set):
        """
        Return the length of the final collection, directly asking redis for the
        count without calling sort
        """
        conn = self.cls.get_connection()

        # we have a sorted set without need to sort, use zcard
        if self._has_sortedsets:
            return conn.zcard(final_set)

        # we have a stored collection, without other filter, use llen
        elif self.stored_key and not self._lazy_collection['sets']\
                and len(self._lazy_collection['intersects']) == 1:

            return conn.llen(final_set)

        # normal call
        return super(ExtendedCollectionManager, self)._collection_length(final_set)

    def sort(self, **parameters):
        """
        Enhance the default sort method to accept a new parameter "by_score", to
        use instead of "by" if you want to sort by the score of a sorted set.
        You must pass to "by_sort" the key of a redis sorted set (or a
        sortedSetField attached to an instance)
        """
        self._sort_by_sortedset = None
        is_sortedset = False
        if parameters.get('by_score'):
            if parameters.get('by'):
                raise ValueError("You can't use `by` and `by_score` in the same "
                                 "call to `sort`.")
            by = parameters.get('by_score', None)
            if isinstance(by, SortedSetField) and getattr(by, '_instance', None):
                by = by.key
            elif not isinstance(by, str):
                by = None

            if by is None:
                raise ValueError('To sort by sorted set, you must pass a '
                                 'SortedSetFied (attached to a model) or a '
                                 'string representing the key of a redis zset '
                                 'to the `by_score` named argument')
            is_sortedset = True
            parameters['by'] = by
            del parameters['by_score']

        else:
            # allow passing a field, not only a field name
            by = parameters.get('by')
            if by and isinstance(by, RedisField):
                parameters['by'] = by.name

        super(ExtendedCollectionManager, self).sort(**parameters)

        if is_sortedset:
            self._sort_by_sortedset = self._sort
            self._sort = None

        return self

    def _zset_to_keys(self, key, values=None, alpha=False):
        """
        Convert a redis sorted set to a list of keys, to be used by sort.
        Each key is on the following format, for each value in the sorted set:
            ramdom_string:value-in-the-sorted-set => score-of-the-value
        The random string is the same for all keys.
        If values is not None, only these values from the sorted set are saved
        as keys.
        If a value in values is not on the sorted set, it's still saved as a key
        but with a default value ('' is alpha is True, else '-inf')
        """
        conn = self.cls.get_connection()
        default = '' if alpha else '-inf'
        if values is None:
            # no values given, we get scores from the whole sorted set
            result = conn.zrange(key, start=0, end=-1, withscores=True)
            values = list(islice(chain.from_iterable(result), 0, None, 2))
        else:
            # we have values, we'll get only their scores

            if isinstance(self.cls.database, PipelineDatabase):
                # if available, use the pipeline of our database to get all
                # scores in one redis call
                with self.cls.database.pipeline(transaction=False) as pipe:
                    for value in values:
                        pipe.zscore(key, value)
                    scores = pipe.execute()
            else:
                # no pipeline, we have to do a call for each value
                scores = []
                for value in values:
                    scores.append(conn.zscore(key, value))

            # combine values and scores in one list
            result = []
            for index, value in enumerate(values):
                score = scores[index]
                if score is None:
                    score = default
                result.append((value, score))

        # create a temporary key for each (value,score) tuple
        base_tmp_key = self._unique_key()
        conn.set(base_tmp_key, 'working...')  # only to "reserve" the main tmp key
        tmp_keys = []
        # use a mapping dict (tmp_key_with_value=>score) to use in mset
        mapping = {}
        for value, score in result:
            tmp_key = '%s:%s' % (base_tmp_key, value)
            tmp_keys.append(tmp_key)
            mapping[tmp_key] = score
        # set all keys in one call
        conn.mset(mapping)

        return base_tmp_key, tmp_keys

    def _prepare_sort_by_score(self, values, sort_options):
        """
        Create the key to sort on the sorted set references in
        self._sort_by_sortedset and adapte sort options
        """
        # create the keys
        base_tmp_key, tmp_keys = self._zset_to_keys(
                                    key=self._sort_by_sortedset['by'],
                                    values=values,
                                    )
        # ask to sort on our new keys
        sort_options['by'] = '%s:*' % base_tmp_key

        # retrieve original sort parameters
        for key in ('desc', 'alpha', 'get', 'store'):
            if key in self._sort_by_sortedset:
                sort_options[key] = self._sort_by_sortedset[key]

        # if we want to get the score with values/values_list
        if sort_options.get('get'):
            try:
                pos = sort_options['get'].index(SORTED_SCORE)
            except:
                pass
            else:
                sort_options['get'][pos] = '%s:*' % base_tmp_key

        return base_tmp_key, tmp_keys

    def _prepare_results(self, results):
        """
        Sort results by score if not done before (faster, if we have no values to
        retrieve, or slice)
        """
        # if we want a result sorted by a score, and if we have a full result
        # (no slice), we can do it know, by creating keys for each values with
        # the sorted set score, and sort on them
        if self._sort_by_sortedset_after and (len(results) > 1 or self._values):
            conn = self.cls.get_connection()

            sort_params = {}
            base_tmp_key, tmp_keys = self._prepare_sort_by_score(results, sort_params)

            # compose the set to sort
            final_set = '%s_final_set' % base_tmp_key
            conn.sadd(final_set, *results)

            # apply the sort
            results = conn.sort(final_set, **sort_params)

            # finally delete all temporary keys
            conn.delete(*(tmp_keys + [final_set, base_tmp_key]))

        if self._store:
            # if store, redis doesn't return result, so don't return anything here
            return

        if self._values and self._values['mode'] != 'flat':
            results = self._to_values(results)

        return super(ExtendedCollectionManager, self)._prepare_results(results)

    def _to_values(self, collection):
        """
        Regroup values in tuples or dicts for each "instance".
        Exemple: Given this result from redis: ['id1', 'name1', 'id2', 'name2']
         tuples: [('id1', 'name1'), ('id2', 'name2')]
         dicts:  [{'id': 'id1', 'name': 'name1'}, {'id': 'id2', 'name': 'name2'}]
        """
        result = zip(*([iter(collection)] * len(self._values['fields']['names'])))
        if self._values['mode'] == 'dicts':
            result = (dict(zip(self._values['fields']['names'], a_result)) for a_result in result)
        return result

    @property
    def _sort_by_sortedset_before(self):
        """
        Return True if we have to sort by set and do the stuff *before* asking
        redis for the sort
        """
        return self._sort_by_sortedset and self._sort_limits and (not self._lazy_collection['pks']
                                                            or self._want_score_value)

    @property
    def _sort_by_sortedset_after(self):
        """
        Return True if we have to sort by set and do the stuff *after* asking
        redis for the sort
        """
        return self._sort_by_sortedset and not self._sort_limits and (not self._lazy_collection['pks']
                                                                or self._want_score_value)

    @property
    def _want_score_value(self):
        """
        Return True if we want the score of the sorted set used to sort in the
        results from values/values_list
        """
        return self._values and SORTED_SCORE in self._values['fields']['names']

    def _prepare_sort_options(self, has_pk):
        """
        Prepare sort options for _values attributes.
        If we manager sort by score after getting the result, we do not want to
        get values from the first sort call, but only from the last one, after
        converting results in zset into keys
        """
        sort_options = super(ExtendedCollectionManager, self)._prepare_sort_options(has_pk)

        if self._values:
            # if we asked for values, we have to use the redis 'sort'
            # command, which is able to return other fields.
            if not sort_options:
                sort_options = {}
            sort_options['get'] = self._values['fields']['keys']

        if self._sort_by_sortedset_after:
            for key in ('get', 'store'):
                if key in self._sort_by_sortedset:
                    del self._sort_by_sortedset[key]
            if sort_options and (not has_pk or self._want_score_value):
                for key in ('get', 'store'):
                    if key in sort_options:
                        self._sort_by_sortedset[key] = sort_options.pop(key)
            if not sort_options:
                sort_options = None
        return sort_options

    def _get_final_set(self, sets, pk, sort_options):
        """
        Add intersects fo sets and call parent's _get_final_set.
        If we have to sort by sorted score, and we have a slice, we have to
        convert the whole sorted set to keys now.
        """
        if self._lazy_collection['intersects']:
            # if the intersect method was called, we had new sets to intersect
            # to the global set of sets.
            # And it there is no real filters, we had the set of the whole
            # collection because we cannot be sure that entries in "intersects"
            # are all real primary keys
            sets = sets[::]
            sets.extend(self._lazy_collection['intersects'])
            if not self._lazy_collection['sets'] and not self.stored_key:
                sets.append(self.cls.get_field('pk').collection_key)

        final_set, keys_to_delete_later = super(ExtendedCollectionManager,
                                    self)._get_final_set(sets, pk, sort_options)

        # if we have a slice and we want to sort by the score of a sorted set,
        # as redis sort command doesn't handle this, we have to create keys for
        # each value of the sorted set and sort on them
        # @antirez, y u don't allow this !!??!!
        if final_set and self._sort_by_sortedset_before:
            # TODO: if we have filters, maybe apply _zet_to_keys to only
            #       intersected values
            base_tmp_key, tmp_keys = self._prepare_sort_by_score(None, sort_options)
            # new keys have to be deleted once the final sort is done
            if not keys_to_delete_later:
                keys_to_delete_later = []
            keys_to_delete_later.append(base_tmp_key)
            keys_to_delete_later += tmp_keys

        return final_set, keys_to_delete_later

    def _add_filters(self, **filters):
        """
        In addition to the normal _add_filters, this one accept RedisField objects
        on the right part of a filter. The value will be fetched from redis when
        calling the collection.
        The filter value can also be a model instance, in which case its PK will
        be fetched when calling the collection, too.
        """
        string_filters = filters.copy()

        for key, value in filters.items():

            is_extended = False

            if isinstance(value, RedisField):
                # we will fetch the value when running the collection
                if (not isinstance(value, SingleValueField)
                    or getattr(value, '_instance', None) is None):
                    raise ValueError('If a field is used as a filter value, it '
                                     'must be a simple value field attached to '
                                     'an instance')
                is_extended = True

            elif isinstance(value, RedisModel):
                # we will fetch the PK when running the collection
                is_extended = True

            if is_extended:
                if self._field_is_pk(key):
                    # create an RawFilter which will be used in _get_pk
                    raw_filter = RawFilter(key, value)
                    self._lazy_collection['pks'].add(raw_filter)
                else:
                    # create an ParsedFilter which will be used in _prepare_sets
                    index, suffix, extra_field_parts = self._parse_filter_key(key)
                    parsed_filter = ParsedFilter(index, suffix, extra_field_parts, value)
                    self._lazy_collection['sets'].append(parsed_filter)

                string_filters.pop(key)

        super(ExtendedCollectionManager, self)._add_filters(**string_filters)

        return self

    def _get_pk(self):
        """
        Override the default _get_pk method to retrieve the real pk value if we
        have a SingleValueField or a RedisModel instead of a real PK value
        """
        pk = super(ExtendedCollectionManager, self)._get_pk()

        if pk is not None and isinstance(pk, RawFilter):
            # We have a RedisModel and we want its pk, or a RedisField
            # (single value) and we want its value
            if isinstance(pk.value, RedisModel):
                pk = pk.value.pk.get()
            elif isinstance(pk.value, SingleValueField):
                pk = pk.value.proxy_get()
            else:
                raise ValueError(u'Invalide filter value for a PK: %s' % pk.value)

        return pk

    def _coerce_fields_parameters(self, fields):
        """
        Used by values and values_list to get the list of fields to use in the
        redis sort command to retrieve fields.
        The result is a dict with two lists:
          - 'names', with wanted field names
          - 'keys', with keys to use in the sort command
        When sorting by score, we allow to retrieve the score in values/values_list.
        For this, just pass SORTED_SCORE (importable from contrib.collection) as
        a name to retrieve.
        If finally the result is not sorted by score, the value for this part
        will be None
        """
        try:
            sorted_score_pos = fields.index(SORTED_SCORE)
        except:
            sorted_score_pos = None
        else:
            fields = list(fields)
            fields.pop(sorted_score_pos)

        final_fields = {'names': [], 'keys': []}
        for field_name in fields:
            if self._field_is_pk(field_name):
                final_fields['names'].append(field_name)
                final_fields['keys'].append('#')
            else:
                if not self.cls.has_field(field_name):
                    raise ValueError("%s if not a valid field to get from collection"
                                     " for %s" % (field_name, self.cls.__name__))
                field = self.cls.get_field(field_name)
                if isinstance(field, MultiValuesField):
                    raise ValueError("It's not possible to get a MultiValuesField"
                                     " from a collection (asked: %s" % field_name)
                final_fields['names'].append(field_name)
                final_fields['keys'].append(field.sort_wildcard)

        if sorted_score_pos is not None:
            final_fields['names'].insert(sorted_score_pos, SORTED_SCORE)
            final_fields['keys'].insert(sorted_score_pos, SORTED_SCORE)

        return final_fields

    def store(self, key=None, ttl=DEFAULT_STORE_TTL):
        """
        Will call the collection and store the result in Redis, and return a new
        collection based on this stored result. Note that only primary keys are
        stored, ie calls to values/values_list are ignored when storing result.
        But choices about instances/values_list are transmited to the new
        collection.
        If no key is given, a new one will be generated.
        The ttl is the time redis will keep the new key. By default its
        DEFAULT_STORE_TTL, which is 60 secondes. You can pass None if you don't
        want expiration.
        """
        old_sort_limits_and_len_mode = None if self._sort_limits is None else self._sort_limits.copy(), self._len_mode
        try:
            self._store = True

            # save sort and values options
            sort_options = None
            if self._sort is not None:
                sort_options = self._sort.copy()
            values = None
            if self._values is not None:
                values = self._values
                self._values = None

            # create a key for storage
            store_key = key or self._unique_key()
            if self._sort is None:
                self._sort = {}
            self._sort['store'] = store_key

            # if filter by pk, but without need to get "values", no redis call is done
            # so force values to get a call to sort (to store result)
            if self._lazy_collection['pks'] and not self._values:
                self.values('pk')

            # call the collection
            self._len_mode = False
            self._collection

            # restore sort and values options
            self._store = False
            self._sort = sort_options
            self._values = values

            # create the new collection
            stored_collection = self.__class__(self.cls)
            stored_collection.from_stored(store_key)

            # apply ttl if needed
            if ttl is not None:
                self.cls.get_connection().expire(store_key, ttl)

            # set choices about instances/values from the current to the new collection
            for attr in ('_instances', '_instances_skip_exist_test', '_values'):
                setattr(stored_collection, attr, deepcopy(getattr(self, attr)))

            # finally return the new collection
            return stored_collection

        finally:
            self._sort_limits, self._len_mode = old_sort_limits_and_len_mode

    def from_stored(self, key):
        """
        Set the current collection as based on a stored one. The key argument
        is the key off the stored collection.
        """
        # only one stored key allowed
        if self.stored_key:
            raise ValueError('This collection is already based on a stored one')

        # prepare the collection
        self.stored_key = key
        self.intersect(_StoredCollection(self.cls.get_connection(), key))
        self.sort(by='nosort')  # keep stored order

        # count the number of results to manage empty result (to not behave like
        # expired key)
        self._stored_len = self.cls.get_connection().llen(key)

        return self

    def stored_key_exists(self):
        """
        Check the existence of the stored key (useful if the collection is based
        on a stored one, to check if the redis key still exists)
        """
        return self.cls.get_connection().exists(self.stored_key)

    def reset_result_type(self):
        """
        Reset the type of values attened for the collection (ie cancel a
        previous "instances" or "values" call)
        """
        self._values = None
        return super(ExtendedCollectionManager, self).reset_result_type()

    def values(self, *fields):
        """
        Ask the collection to return a list of dict of given fields for each
        instance found in the collection.
        If no fields are given, all "simple value" fields are used.
        """
        if not fields:
            fields = self._get_simple_fields()

        fields = self._coerce_fields_parameters(fields)

        self._instances = False
        self._values = {'fields': fields, 'mode': 'dicts'}
        return self

    def values_list(self, *fields, **kwargs):
        """
        Ask the collection to return a list of tuples of given fields (in the
        given order) for each instance found in the collection.
        If 'flat=True' is passed, the resulting list will be flat, ie without
        tuple. It's a valid kwarg only if only one field is given.
        If no fields are given, all "simple value" fields are used.
        """
        flat = kwargs.pop('flat', False)
        if kwargs:
            raise ValueError('Unexpected keyword arguments for the values method: %s'
                             % list(kwargs))

        if not fields:
            fields = self._get_simple_fields()

        if flat and len(fields) > 1:
            raise ValueError("'flat' is not valid when values is called with more than one field.")

        fields = self._coerce_fields_parameters(fields)

        self._instances = False
        self._values = {'fields': fields, 'mode': 'flat' if flat else 'tuples'}
        return self


class _StoredCollection(object):
    """
    Simple object to store the key of a stored collection, to be used in
    ExtendedCollectionManager based on a stored collection.
    The stored key is a list, so it's managed as a ListField (but we only need
    its key, and lmembers if no scripting)
    """
    def __init__(self, connection, key):
        self.connection = connection
        self.key = key

    def lmembers(self):
        """
        Return the list of all members of the list, used by _list_to_set if
        no scripting
        """
        return self.connection.lrange(self.key, 0, -1)



# -*- coding:utf-8 -*-
from __future__ import unicode_literals
from future.builtins import object
from collections import namedtuple


from limpyd.utils import unique_key
from limpyd.exceptions import *
from limpyd.fields import MultiValuesField


ParsedFilter = namedtuple('ParsedFilter', ['index', 'suffix', 'extra_field_parts', 'value'])

NONE_SLICE = slice(None, None, None)


class CollectionManager(object):
    """
    Retrieve objects collection, optionnaly slice and order it.
    CollectionManager is lazy: it will call redis only when evaluating it
    (iterating, slicing, forcing as list...)

    API:
    MyModel.collection(**filters) => return the whole collection, eventually filtered.
    MyModel.collection().sort(by='field') => return the collection sorted.
    MyModel.collection().sort(by='field')[:10] => slice the sorted collection.
    MyModel.collection().instances() => return the instances

    Note:
    Slicing a collection will force a sort.
    """

    _accepted_key_types = {'set'}  # Type of keys indexes are allowed to return

    def __init__(self, cls):
        self.cls = cls
        self._lazy_collection = {  # Store infos to make the requested collection
            'sets': [],  # store sets to use (we'll intersect them)
            'pks': set(),  # store special filter on pk
        }
        self._instances = False  # True when instances are asked
                                 # instead of raw pks
        self._instances_skip_exist_test = False  # If True will return instances
                                                 # without testing if pk exist
        self._sort = None  # Will store sorting parameters
        self._sort_limits = None  # Will store slice parameters (start and num)
        self._len = None  # Store the result of the final collection, to avoid
                          # having to compute the whole thing twice when doing
                          # `list(Model.collection)` (a `list` will call
                          # __iter__ AND __len__)
        self._len_mode = True   # Set to True if the __len__ method is directly
                                # called (to avoid some useless computations)
                                # True by default to manage the __iter__ + __len__
                                # case, specifically set to False in other cases

        self.scripted_slicing = True

    def __iter__(self):
        old_sort_limits_and_len_mode = None if self._sort_limits is None else self._sort_limits.copy(), self._len_mode
        try:
            self._len_mode = False
            return self._collection.__iter__()
        finally:
            self._sort_limits, self._len_mode = old_sort_limits_and_len_mode

    def _optimize_slice(self, the_slice, can_reverse):
        """

        Parameters
        ----------
        the_slice: slice
            The python slice to optimize
        can_reverse: bool
            If we can ask redis to reverse the data before slicing.

        Returns
        -------
        tuple
            Four elements:
            - optimized: bool
                If the slice is optimized
            - start: int or None
                start item to pass to the redis command
                If None, we want from the very start
            - count: int or None
                number of items to ask the redi command
                If None, we want everything from the start
            - rev: bool
                If we should ask redis to work the reverse way
            - python_slice
                The slice to be applied on the python side

        """
        step = 1 if the_slice.step in (1, None) else the_slice.step
        start = (the_slice.start or 0) if step > 0 else the_slice.start
        stop = the_slice.stop

        if start is not None and stop is not None:

            # cases where we know the result is always empty
            if step > 0 and (0 <= stop <= start or stop <= start < 0):
                return True, None, None, False, NONE_SLICE

            if step < 0 and (0 <= start <= stop or start <= stop < 0):
                return True, None, None, False, NONE_SLICE

            # if start and stop are on the same side, we can optimize

            # simplest case: start and stop are positive
            if start >= 0 and stop > 0:
                if step < 0:
                    start, stop = stop + 1, start + 1
                return True, start, stop - start, False, slice(None, None, step if step != 1 else None)

            # more complicated case: start and stop are negative
            if start < 0 and stop < 0 and can_reverse:
                if step < 0:
                    start, stop = stop + 1, start + 1
                step = -step
                return True, - stop, stop - start, True, slice(None, None, step if step != 1 else None)

        # for positive step, if we have a positive start, and no end or negative one, we have to retrieve all from
        # the start and slice here
        if step > 0 and (start is not None and start >= 0) and (stop is None or stop < 0):
            return True, start, None, False, slice(None, stop, step)

        # the reverse for negative step
        if step < 0 and (start is None or start < 0) and (stop is not None and stop >= 0):
            return True, stop + 1, None, False, slice(start, None, step)

        # in all other cases we recover the whole collection and slice it in python
        return False, None, None, False, slice(start, stop, step)

    def __getitem__(self, arg):
        old_sort_limits_and_len_mode = None if self._sort_limits is None else self._sort_limits.copy(), self._len_mode
        old_sort = None if self._sort is None else self._sort.copy()
        try:
            self._len_mode = False
            self._sort_limits = {}
            self._optimized_slicing = False  # just to store the fact that the call was optimized
            if isinstance(arg, slice):
                # A slice has been requested
                # We try to reduce the data to get back from redis
                # when it's possible depending of the slice arguments.
                # We use `self._sort_limits` to tell redis how to slice what
                # it'll got from the `sort` command. We may also reverse
                # `_sort['desc']`(the set will be read in reverse way as we wanted
                # by redis).
                # At the end we return the collection (a sliced collection
                # is no more chainable, so we do not return `self`

                optimized, start, count, rev, python_slice = self._optimize_slice(
                    arg,
                    can_reverse=True if not self._sort else self._sort.get('by', '') != 'nosort'
                )

                self._optimized_slicing = optimized

                if optimized and start is None and count is None and python_slice == NONE_SLICE:
                    return []

                if start is not None or count is not None:
                    self._sort_limits['start'] = start or 0
                    self._sort_limits['num'] = -1 if count is None else count

                if rev:
                    if self._sort is None: self._sort = {}
                    self._sort['desc'] = not self._sort.get('desc', False)

                return self._collection[python_slice]

            else:
                # A single item has been requested
                # Nevertheless, use the redis pagination, to minimize
                # data transfer and use the fast redis offset system
                start = arg
                self._sort_limits['num'] = 1  # one element
                self._optimized_slicing = True
                if start >= 0:
                    self._sort_limits['start'] = start
                    return self._collection[0]
                else:
                    # we sort the result in the reverse way, mark the final result as
                    # reversed to re-reverse it at the end
                    if self._sort is None: self._sort = {}
                    self._sort['desc'] = not self._sort.get('desc', False)
                    self._sort_limits['start'] = - start - 1
                    return self._collection[0]
        finally:
            self._sort_limits, self._len_mode = old_sort_limits_and_len_mode
            self._sort = old_sort

    def _get_pk(self):
        """
        Return None if we don't have any filter on a pk, the pk if we have one,
        or raise a ValueError if we have more than one.
        For internal use only.
        """
        pk = None
        if self._lazy_collection['pks']:
            if len(self._lazy_collection['pks']) > 1:
                raise ValueError('Too much pks !')
            pk = list(self._lazy_collection['pks'])[0]
        return pk

    def _prepare_sort_options(self, has_pk):
        """
        Prepare "sort" options to use when calling the collection, depending
        on "_sort" and "_sort_limits" attributes
        """
        sort_options = {}
        if self._sort is not None and not has_pk:
            sort_options.update(self._sort)
        if self._sort_limits is not None:
            if 'start' in self._sort_limits and 'num' not in self._sort_limits:
                self._sort_limits['num'] = -1
            elif 'num' in self._sort_limits and 'start' not in self._sort_limits:
                self._sort_limits['start'] = 0
            sort_options.update(self._sort_limits)
        if not sort_options and self._sort is None:
            sort_options = None
        return sort_options

    @property
    def _collection(self):
        """
        Effectively retrieve data according to lazy_collection.
        """
        old_sort_limits_and_len_mode = None if self._sort_limits is None else self._sort_limits.copy(), self._len_mode
        try:  # try block to always reset the _sort_limits in the "finally" part

            conn = self.cls.get_connection()
            self._len = 0

            # The collection fails (empty) if more than one pk or if the only one
            # doesn't exists
            try:
                pk = self._get_pk()
            except ValueError:
                return []
            else:
                if pk is not None and not self.cls.get_field('pk').exists(pk):
                    return []

            # Prepare options and final set to get/sort
            sort_options = self._prepare_sort_options(bool(pk))

            final_set, keys_to_delete = self._get_final_set(
                                                self._lazy_collection['sets'],
                                                pk, sort_options)

            if self._len_mode:
                if final_set is None:
                    if pk and not self._lazy_collection['sets']:
                        # we have a only pk
                        self._len = 1
                    else:
                        # we have nothing
                        self._len = 0
                else:
                    self._len = self._collection_length(final_set)
                    if keys_to_delete:
                        conn.delete(*keys_to_delete)

                # return nothing
                return

            else:

                # fill the collection
                if final_set is None:
                    if pk and not self._lazy_collection['sets']:
                        # we have a pk without other sets, and no
                        # needs to get values so we can simply return the pk
                        collection = {pk}
                    else:
                        # we have nothing
                        collection = {}
                else:
                    # compute the sets and call redis te retrieve wanted values
                    collection = self._final_redis_call(final_set, sort_options)
                    if keys_to_delete:
                        conn.delete(*keys_to_delete)

                # Format return values if needed
                return self._prepare_results(collection)
        finally:
            self._sort_limits, self._len_mode = old_sort_limits_and_len_mode

    def _final_redis_call(self, final_set, sort_options):
        """
        The final redis call to obtain the values to return from the "final_set"
        with some sort options.
        """

        conn = self.cls.get_connection()
        if sort_options is not None:
            # a sort, or values, call the SORT command on the set
            return conn.sort(final_set, **sort_options)
        else:
            # no sort, nor values, simply return the full set
            return conn.smembers(final_set)

    def _collection_length(self, final_set):
        """
        Return the length of the final collection, directly asking redis for the
        count without calling sort
        """
        return self.cls.get_connection().scard(final_set)

    def _to_instances(self, pks):
        """
        Returns a list of instances for each given pk, respecting the condition
        about checking or not if a pk exists.
        """
        # we want instances, so create an object for each pk, without
        # checking for pk existence if asked
        meth = self.cls.lazy_connect if self._instances_skip_exist_test else self.cls
        return [meth(pk) for pk in pks]

    def _prepare_results(self, results):
        """
        Called in _collection to prepare results from redis before returning
        them.
        """
        if self._instances:
            results = self._to_instances(results)
        else:
            results = list(results)

        # cache the len for future use
        self._len = len(results)

        return results

    def _prepare_sets(self, sets):
        """
        Return all sets in self._lazy_collection['sets'] to be ready to be used
        to intersect them. Called by _get_final_set, to use in subclasses.
        Must return a tuple with a set of redis set keys, and another with
        new temporary keys to drop at the end of _get_final_set
        """

        final_sets = set()
        tmp_keys = set()

        for set_ in sets:
            if isinstance(set_, str):
                final_sets.add(set_)
            elif isinstance(set_, ParsedFilter):

                for index_key, key_type, is_tmp in set_.index.get_filtered_keys(
                            set_.suffix,
                            accepted_key_types=self._accepted_key_types,
                            *(set_.extra_field_parts + [set_.value])
                        ):
                    if key_type not in self._accepted_key_types:
                        raise ValueError('The index key returned by the index %s is not valid' % (
                            set_.index.__class__.__name__
                        ))
                    final_sets.add(index_key)
                    if is_tmp:
                        tmp_keys.add(index_key)
            else:
                raise ValueError('Invalid filter type')

        return final_sets, tmp_keys

    def _get_final_set(self, sets, pk, sort_options):
        """
        Called by _collection to get the final set to work on. Return the name
        of the set to use, and a list of keys to delete once the collection is
        really called (in case of a computed set based on multiple ones)
        """
        conn = self.cls.get_connection()
        all_sets = set()
        tmp_keys = set()

        if pk is not None and not sets and not (sort_options and sort_options.get('get')):
            # no final set if only a pk without values to retrieve
            return (None, False)

        elif sets or pk:
            if sets:
                new_sets, new_tmp_keys = self._prepare_sets(sets)
                all_sets.update(new_sets)
                tmp_keys.update(new_tmp_keys)
            if pk is not None:
                # create a set with the pk to do intersection (and to pass it to
                # the store command to retrieve values if needed)
                tmp_key = self._unique_key()
                conn.sadd(tmp_key, pk)
                all_sets.add(tmp_key)
                tmp_keys.add(tmp_key)

        else:
            # no sets or pk, use the whole collection instead
            all_sets.add(self.cls.get_field('pk').collection_key)

        if not all_sets:
            delete_set_later = False
            final_set = None
        elif len(all_sets) == 1:
            # if we have only one set, we  delete the set after calling
            # collection only if it's a temporary one, and we do not delete
            # it right now
            final_set = all_sets.pop()
            if final_set in tmp_keys:
                delete_set_later = True
                tmp_keys.remove(final_set)
            else:
                delete_set_later = False
        else:
            # more than one set, do an intersection on all of them in a new key
            # that will must be deleted once the collection is called.
            delete_set_later = True
            final_set = self._combine_sets(all_sets, self._unique_key())

        if tmp_keys:
            conn.delete(*tmp_keys)

        # return the final set to work on, and a flag if we later need to delete it
        return (final_set, [final_set] if delete_set_later else None)

    def _combine_sets(self, sets, final_set):
        """
        Given a list of set, combine them to create the final set that will be
        used to make the final redis call.
        """
        self.cls.get_connection().sinterstore(final_set, list(sets))
        return final_set

    def __call__(self, **filters):
        return self._add_filters(**filters)

    def _field_is_pk(self, field_name):
        """Check if the given name is the pk field, suffixed or not with "__eq" """
        if self.cls._field_is_pk(field_name):
            return True
        if field_name.endswith('__eq') and self.cls._field_is_pk(field_name[:-4]):
            return True
        return False

    def _parse_filter_key(self, key):
        # Each key can have optional subpath
        # We pass it as args to the field, which is responsable
        # from handling them
        # We only manage here the suffix handled by a filter

        key_path = key.split('__')
        field_name = key_path.pop(0)
        field = self.cls.get_field(field_name)

        if not field.indexable:
            raise ImplementationError(
                'Field %s.%s is not indexable' % (
                    field._model.__name__, field.name
                )
            )

        other_field_parts = key_path[:field._field_parts - 1]

        if len(other_field_parts) + 1 != field._field_parts:
            raise ImplementationError(
                'Unexpected number of parts in filter %s for field %s.%s' % (
                    key, field._model.__name__, field.name
                )
            )

        rest = key_path[field._field_parts - 1:]
        index_suffix = None if not rest else '__'.join(rest)
        index_to_use = None
        for index in field._indexes:
            if index.can_handle_suffix(index_suffix):
                index_to_use = index
                break

        if not index_to_use:
            raise ImplementationError(
                'No index found to manage filter %s for field %s.%s' % (
                    key, field._model.__name__, field.name
                )
            )

        return index_to_use, index_suffix, other_field_parts

    def _add_filters(self, **filters):
        """Define self._lazy_collection according to filters."""
        for key, value in filters.items():
            if self._field_is_pk(key):
                pk = self.cls.get_field('pk').normalize(value)
                self._lazy_collection['pks'].add(pk)
            else:
                # store the info to call the index later, in ``_prepare_sets``
                # (to avoid doing extra work if the collection is never called)
                index, suffix, extra_field_parts = self._parse_filter_key(key)
                parsed_filter = ParsedFilter(index, suffix, extra_field_parts, value)
                self._lazy_collection['sets'].append(parsed_filter)

        return self

    def __len__(self):
        if self._len is None:
            if not self._len_mode:
                self._len = self._collection.__len__()
            else:
                self._collection
        return self._len

    def __repr__(self):
        old_sort_limits_and_len_mode = None if self._sort_limits is None else self._sort_limits.copy(), self._len_mode
        try:
            self._len_mode = False
            return self._collection.__repr__()
        finally:
            self._sort_limits, self._len_mode = old_sort_limits_and_len_mode

    def instances(self, skip_exist_test=False):
        """
        Ask the collection to return a list of instances.
        If skip_exist_test is set to True, the instances returned by the
        collection won't have their primary key checked for existence.
        """
        self.reset_result_type()
        self._instances = True
        self._instances_skip_exist_test = skip_exist_test
        return self

    def _get_simple_fields(self):
        """
        Return a list of the names of all fields that handle simple values
        (StringField or InstanceHashField), that redis can use to return values via
        the sort command (so, exclude all fields based on MultiValuesField)
        """
        fields = []
        for field_name in self.cls._fields:
            field = self.cls.get_field(field_name)
            if not isinstance(field, MultiValuesField):
                fields.append(field_name)
        return fields

    def primary_keys(self):
        """
        Ask the collection to return a list of primary keys. It's the default
        but if `instances`, `values` or `values_list` was previously called,
        a call to `primary_keys` restore this default behaviour.
        """
        self.reset_result_type()
        return self

    def reset_result_type(self):
        """
        Reset the type of values attened for the collection (ie cancel a
        previous "instances" call)
        """
        self._instances = False

    def _coerce_by_parameter(self, parameters):
        if "by" in parameters:
            by = parameters['by']
            # Manage desc option
            if by.startswith('-'):
                parameters['desc'] = True
                by = by[1:]
            if self.cls._field_is_pk(by):
                # don't use field, the final set, which contains pks, will be sorted directly
                del parameters['by']
            elif self.cls.has_field(by):
                # if we have a field, use its redis wildcard form
                field = self.cls.get_field(by)
                parameters['by'] = field.sort_wildcard
        return parameters

    def sort(self, **parameters):
        """
        Parameters:
        `by`: pass either a field name or a wildcard string to sort on
              prefix with `-` to make a desc sort.
        `alpha`: set it to True to sort lexicographically instead of numerically.
        """
        parameters = self._coerce_by_parameter(parameters)
        self._sort = parameters
        return self

    def _unique_key(self):
        """
        Create a unique key.
        """
        return unique_key(self.cls.get_connection())



# -*- coding:utf-8 -*-
from __future__ import unicode_literals
from future.utils import iteritems, iterkeys
from future.utils import with_metaclass

from logging import getLogger
from copy import copy
import inspect
import threading

from limpyd.fields import *
from limpyd.utils import make_key
from limpyd.exceptions import *
from limpyd.database import RedisDatabase
from limpyd.collection import CollectionManager

__all__ = ['RedisModel', ]

log = getLogger(__name__)
threadlocal = threading.local()


class MetaRedisModel(MetaRedisProxy):
    """
    We make invisible for user that fields were class properties
    """
    def __new__(mcs, name, base, attrs):

        it = super(MetaRedisModel, mcs).__new__(mcs, name, base, attrs)
        it.abstract = attrs.get('abstract', False)

        if not it.abstract:
            if not hasattr(it, 'database') or not isinstance(it.database, RedisDatabase):
                raise ImplementationError(
                    'You must define a database for the model %s' % name)
            if not getattr(it, 'namespace', None):
                it.namespace = ''
            it._name = ':'.join((it.namespace, name.lower()))

            # Get the caller (next frame in the stack), with it's source file and line number
            # It will be used in ``database._add_model``, if a model with the same name
            # already exists, to check if it's the same one.
            it._creation_source = inspect.stack()[1][1:3]
            it_in_db = it.database._add_model(it)
            # If the returned model is not the same, it's the one from the database
            # we already added for this name, so we use it
            if it_in_db is not it:
                return it_in_db

        # init (or get from parents) lists of redis fields
        _fields = list(it._fields) if hasattr(it, '_fields') else []
        _instancehash_fields = list(it._instancehash_fields) if hasattr(it,
                                                                    '_instancehash_fields') else []

        # Did we have already pk field ?
        pk_field = getattr(it, '_redis_attr_pk', None)

        # First loop on new attributes for this class to find fields and
        # primary key, and validate the eventually found PKField
        own_fields = []

        # limit to redis fields
        redis_fields = [(k, v) for k, v in attrs.items() if not k.startswith('_')
                                                            and isinstance(v, RedisField)]
        # and keep them by declaration order
        redis_fields = [(k, v) for k, v in sorted(redis_fields, key=lambda f: f[1]._creation_order)]

        for (field_name, field) in redis_fields:
            field.name = field_name  # each field must know its name
            if isinstance(field, PKField):
                # Check and save the primary key
                if pk_field:
                    # If a PKField already exists, remove the previously auto-added
                    if pk_field._auto_added:
                        _fields.remove(pk_field.name)
                    else:
                        raise ImplementationError(
                            'Only one PKField field is allowed on %s' % name)
                pk_field = field
            own_fields.append(field)

        # We have to store the name of the class on which a field is attached
        # to compute needed redis keys.
        # For this, a model and its subclasses must not share fields, so we
        # copy existing ones (from the parent class) to the current class.
        for field_name in _fields:
            key = "_redis_attr_%s" % field_name
            field = getattr(it, key)
            ownfield = copy(field)
            ownfield._attach_to_model(it)
            setattr(it, key, ownfield)

        # Auto create missing primary key (it will always be called in RedisModel)
        if not pk_field:
            pk_field = AutoPKField()
            pk_field._auto_added = True
            own_fields.append(pk_field)

        # Loop on new fields to prepare them
        for field in own_fields:
            # remove the original field from the class
            if field.name in attrs:
                delattr(it, field.name)
            # and attach it to the model with its prefixed name
            field._attach_to_model(it)
            _fields.append(field.name)
            setattr(it, "_redis_attr_%s" % field.name, field)
            # save InstanceHashFields in a special list
            if isinstance(field, InstanceHashField):
                _instancehash_fields.append(field.name)

        # keep the pk as first field
        _fields.remove(pk_field.name)
        _fields.insert(0, pk_field.name)

        # Save usefull attributes on the final model
        it._fields = _fields
        it._instancehash_fields = _instancehash_fields
        if pk_field.name != 'pk':
            it._redis_attr_pk = getattr(it, "_redis_attr_%s" % pk_field.name)

        return it


class RedisModel(with_metaclass(MetaRedisModel, RedisProxyCommand)):
    """
    Base redis model.
    """

    namespace = None  # all models in an app may have the same namespace
    lockable = True
    abstract = True
    collection_manager = CollectionManager
    DoesNotExist = DoesNotExist
    default_indexes = None

    available_getters = ('hmget', 'hgetall', 'hkeys', 'hvals', 'hlen')
    available_modifiers = ('hmset', 'hdel')

    def __init__(self, *args, **kwargs):
        """
        Init or retrieve an object storage in Redis.

        Here whats init manages:
        - no args, no kwargs: just instanciate in a python way, no connection to
          redis
        - some kwargs == instanciate, connect, and set the properties received
        - one arg == get from pk
        """
        self.lockable = self.__class__.lockable

        # set to True when the instance's PK will be tested for existence in redis
        self._connected = False

        # --- Meta stuff
        # Put back the fields with the original names
        for attr_name in self._fields:
            attr = getattr(self, "_redis_attr_%s" % attr_name)
            # Copy it, to avoid sharing fields between model instances
            newattr = copy(attr)
            newattr._attach_to_instance(self)
            setattr(self, attr_name, newattr)

        # The `pk` field always exists, even if the real pk has another name
        pk_field_name = getattr(self, "_redis_attr_pk").name
        if pk_field_name != 'pk':
            self.pk = getattr(self, pk_field_name)
        # Cache of the pk value
        self._pk = None

        # change the get_field method to use the instance related one instead
        # of the classmethod
        self.get_field = self.get_instance_field

        # Validate arguments
        if len(args) > 0 and len(kwargs) > 0:
            raise ValueError('Cannot use args and kwargs to instanciate.')

        # save name of fields given when creating the instance, to avoid setting
        # them in the `_set_default` method
        self._init_fields = set()

        # --- Instanciate new from kwargs
        if len(kwargs) > 0:
            # First check unique fields
            # (More robust than trying to manage a "pseudotransaction", as
            # redis do not has "real" transactions)
            # Here we do not set anything, in case one unique field fails
            kwargs_pk_field_name = None
            for field_name, value in iteritems(kwargs):
                if self._field_is_pk(field_name):
                    if kwargs_pk_field_name:
                        raise ValueError(u'You cannot pass two values for the '
                                           'primary key (pk and %s)' % pk_field_name)
                    kwargs_pk_field_name = field_name
                    # always use the real field name, not always pk
                    field_name = pk_field_name
                if not self.has_field(field_name):
                    raise ValueError(u"`%s` is not a valid field name "
                                      "for `%s`." % (field_name, self.__class__.__name__))
                field = self.get_field(field_name)
                if field.unique:
                    field.check_uniqueness(value)
                self._init_fields.add(field_name)

            # Do instanciate, starting by the pk and respecting fields order
            if kwargs_pk_field_name:
                self.pk.set(kwargs[kwargs_pk_field_name])
            for field_name in self._fields:
                if field_name not in kwargs or self._field_is_pk(field_name):
                    continue
                field = self.get_field(field_name)
                field.proxy_set(kwargs[field_name])

        # --- Instanciate from DB
        if len(args) == 1:
            self._pk = self.pk.normalize(args[0])
            self.connect()

    @classmethod
    def get_default_indexes(cls):
        if cls.default_indexes is not None:
            return cls.default_indexes
        return cls.database.get_default_indexes()

    def connect(self):
        """
        Connect the instance to redis by checking the existence of its primary
        key. Do nothing if already connected.
        """
        if self.connected:
            return
        pk = self._pk
        if self.exists(pk=pk):
            self._connected = True
        else:
            self._pk = None
            self._connected = False
            raise DoesNotExist("No %s found with pk %s" % (self.__class__.__name__, pk))

    @classmethod
    def lazy_connect(cls, pk):
        """
        Create an object, setting its primary key without testing it. So the
        instance is not connected
        """
        instance = cls()
        instance._pk = instance.pk.normalize(pk)
        instance._connected = False
        return instance

    @property
    def connected(self):
        """
        A property to check if the model is connected to redis (ie if it as a
        primary key checked for existence)
        """
        return self._connected

    @classmethod
    def use_database(cls, database):
        """
        Transfert the current model to the new database. Move subclasses to the
        new database two if they actually share the same one (so it's easy to
        call use_database on an abstract model to use the new database for all
        subclasses)
        """
        return database._use_for_model(cls)

    @classmethod
    def has_field(cls, field_name):
        """
        Return True if the given field name is an allowed field for this model
        """
        return field_name == 'pk' or field_name in cls._fields

    @classmethod
    def get_class_field(cls, field_name):
        """
        Return the field object with the given name (for the class, the fields
        are in the "_redis_attr_%s" form)
        """
        if not cls.has_field(field_name):
            raise AttributeError('"%s" is not a field for the model "%s"' %
                                 (field_name, cls.__name__))

        field = getattr(cls, '_redis_attr_%s' % field_name)

        return field

    # at the class level, we use get_class_field to get a field
    # but in __init__, we update it to use get_instance_field
    get_field = get_class_field

    def get_instance_field(self, field_name):
        """
        Return the field object with the given name (works for a bound instance)
        """
        if not self.has_field(field_name):
            raise AttributeError('"%s" is not a field for the model "%s"' %
                                 (field_name, self.__class__.__name__))

        field = getattr(self, field_name)

        return field

    def _set_pk(self, value):
        """
        Use the given value as the instance's primary key, if it doesn't have
        one yet (it must be used only for new instances). Then save default values.
        """
        if self._pk:
            raise ImplementationError('Something wrong happened, the PK was already set !')
        self._pk = value
        self._connected = True
        # Default must be set only at first initialization
        self._set_defaults()

    def _set_defaults(self):
        """
        Set default values to fields. We assume that they are not yet populated
        as this method is called just after creation of a new pk.
        """
        for field_name in self._fields:
            if field_name in self._init_fields:
                continue
            field = self.get_field(field_name)
            if hasattr(field, "default"):
                field.proxy_set(field.default)

    @classmethod
    def collection(cls, manager=None, **filters):
        if not manager:
            manager = cls.collection_manager
        collection = manager(cls)
        return collection(**filters)

    @classmethod
    def instances(cls, **filters):
        # FIXME Keep as shortcut or remove for clearer API?
        return cls.collection(**filters).instances()

    @classmethod
    def _field_is_pk(cls, name):
        """
        Check if the given field is the one from the primary key.
        It can be the plain "pk" name, or the real pk field name
        """
        return name in ('pk', cls.get_field('pk').name)

    @classmethod
    def exists(cls, **kwargs):
        """
        A model with the values defined by kwargs exists in db?

        `kwargs` are mandatory.
        """
        if not kwargs:
            raise ValueError(u"`Exists` method requires at least one kwarg.")

        # special case to check for a simple pk
        if len(kwargs) == 1 and cls._field_is_pk(list(kwargs.keys())[0]):
            return cls.get_field('pk').exists(list(kwargs.values())[0])

        # get only the first element of the unsorted collection (the fastest)
        try:
            cls.collection(**kwargs).sort(by='nosort')[0]
        except IndexError:
            return False
        else:
            return True

    @classmethod
    def get(cls, *args, **kwargs):
        """
        Retrieve one instance from db according to given kwargs.

        Optionnaly, one arg could be used to retrieve it from pk.
        """
        if len(args) == 1:  # Guess it's a pk
            pk = args[0]
        elif kwargs:
            # special case to check for a simple pk
            if len(kwargs) == 1 and cls._field_is_pk(list(kwargs.keys())[0]):
                pk = list(kwargs.values())[0]
            else:  # case with many filters
                result = cls.collection(**kwargs).sort(by='nosort')
                if len(result) == 0:
                    raise DoesNotExist(u"No object matching filter: %s" % kwargs)
                elif len(result) > 1:
                    raise ValueError(u"More than one object matching filter: %s" % kwargs)
                else:
                    try:
                        pk = result[0]
                    except IndexError:
                        # object was deleted between the `len` check and now
                        raise DoesNotExist(u"No object matching filter: %s" % kwargs)

        else:
            raise ValueError("Invalid `get` usage with args %s and kwargs %s" % (args, kwargs))
        return cls(pk)

    @classmethod
    def get_or_connect(cls, **kwargs):
        """
        Try to retrieve an object in db, and create it if it does not exist.
        """
        try:
            inst = cls.get(**kwargs)
            created = False
        except DoesNotExist:
            inst = cls(**kwargs)
            created = True
        except Exception:
            raise
        return inst, created

    @classmethod
    def make_key(cls, *args):
        return make_key(*args)

    # --- Hash management
    @property
    def key(self):
        return self.make_key(
            self._name,
            self.pk.get(),
            "hash",
        )

    @classmethod
    def sort_wildcard(cls):
        """
        Used to sort Hashfield. See Hashfield.sort_widlcard.
        """
        return cls.make_key(
            cls._name,
            "*",
            "hash",
        )

    def hmget(self, *args):
        """
        This command on the model allow getting many instancehash fields with only
        one redis call. You must pass hash name to retrieve as arguments.
        """
        if args and not any(arg in self._instancehash_fields for arg in args):
            raise ValueError("Only InstanceHashField can be used here.")

        return self._call_command('hmget', args)

    def hmset(self, **kwargs):
        """
        This command on the model allow setting many instancehash fields with only
        one redis call. You must pass kwargs with field names as keys, with
        their value.
        """
        if kwargs and not any(kwarg in self._instancehash_fields for kwarg in iterkeys(kwargs)):
            raise ValueError("Only InstanceHashField can be used here.")

        indexed = []

        # main try block to revert indexes if something fail
        try:

            # Set indexes for indexable fields.
            for field_name, value in iteritems(kwargs):
                field = self.get_field(field_name)
                if field.indexable:
                    indexed.append(field)
                    field.deindex()
                    field.index(value)

            # Call redis (waits for a dict)
            result = self._call_command('hmset', kwargs)

            return result

        except:
            # We revert indexes previously set if we have an exception, then
            # really raise the error
            for field in indexed:
                field._rollback_indexes()
            raise
        finally:
            for field in indexed:
                field._reset_indexes_caches()

    def hdel(self, *args):
        """
        This command on the model allow deleting many instancehash fields with
        only one redis call. You must pass hash names to retrieve as arguments
        """
        if args and not any(arg in self._instancehash_fields for arg in args):
            raise ValueError("Only InstanceHashField can be used here.")

        # Set indexes for indexable fields.
        for field_name in args:
            field = self.get_field(field_name)
            if field.indexable:
                field.deindex()

        # Return the number of fields really deleted
        return self._call_command('hdel', *args)

    def delete(self):
        """
        Delete the instance from redis storage.
        """
        # Delete each field
        for field_name in self._fields:
            field = self.get_field(field_name)
            if not isinstance(field, PKField):
                # pk has no stored key
                field.delete()
        # Remove the pk from the model collection
        self.connection.srem(self.get_field('pk').collection_key, self._pk)
        # Deactivate the instance
        delattr(self, "_pk")

    @classmethod
    def _thread_lock_storage(cls):
        """
        We mark each locked field in a thread, to allow other operations in the
        same thread on the same field (for the same instance or others). This
        way, operations within the lock, in the same thread, can bypass it (the
        whole set of operations must be locked)
        """
        if not hasattr(threadlocal, 'limpyd_locked_fields'):
            threadlocal.limpyd_locked_fields = {}
        if cls._name not in threadlocal.limpyd_locked_fields:
            threadlocal.limpyd_locked_fields[cls._name] = set()
        return threadlocal.limpyd_locked_fields[cls._name]

    @classmethod
    def _mark_field_as_locked(cls, field):
        cls._thread_lock_storage().add(field.name)

    @classmethod
    def _unmark_field_as_locked(cls, field):
        if field.name in cls._thread_lock_storage():
            cls._thread_lock_storage().remove(field.name)

    @classmethod
    def _is_field_locked(cls, field):
        return field.name in cls._thread_lock_storage()

    def scan_keys(self, count=None):
        """Iter on all the key related to the current instance fields, using redis SCAN command

        Parameters
        ----------
        count: int, default to None (redis uses 10)
            Hint for redis about the number of expected result

        Yields
        -------
        str
            All keys found by the scan, one by one. A key can be returned multiple times, it's
            related to the way the SCAN command works in redis.

        """

        pattern = self.make_key(
            self._name,
            self.pk.get(),
            '*'
        )

        return self.database.scan_keys(pattern, count)

    @classmethod
    def scan_model_keys(cls, count=None):
        """Iter on all the key related to the current model, using redis SCAN command

        Parameters
        ----------
        count: int, default to None (redis uses 10)
            Hint for redis about the number of expected result

        Yields
        -------
        str
            All keys found by the scan, one by one. A key can be returned multiple times, it's
            related to the way the SCAN command works in redis.

        """

        pattern = cls.make_key(
            cls._name,
            "*",
        )

        return cls.database.scan_keys(pattern, count)



# -*- coding:utf-8 -*-
from __future__ import unicode_literals

from limpyd.indexes import BaseIndex, NumberRangeIndex, TextRangeIndex
from limpyd.utils import cached_property


class MultiIndexes(BaseIndex):
    """An index that is a proxy to many ones

    This must not be used directly as a class, but a new index class must be
    created by using the ``create`` class method

    Attributes
    ----------
    index_classes: list
        The index classes composing this multi-indexes class
    key: str
        A key to avoid collision with another index/multi-index
        that will be passed to a field.

    Examples
    --------

        >>> multi_index = MultiIndexes.compose([MyIndex, MyOtherIndex])
        >>> class MyModel(RedisModel):
        ...     field = StringField(indexes=[multi_index])

    """
    index_classes = []

    @classmethod
    def compose(cls, index_classes, key=None, transform=None, name=None):
        """Create a new class with the given index classes

        Parameters
        -----------
        index_classes: list
            The list of index classes to be used in the multi-index class to create
        name: str
            The name of the new multi-index class. If not set, it will be the same
            as the current class
        key: str
            A key to augment the default key of each index, to avoid collision.
        transform: callable
            None by default, can be set to a function that will transform the value to be indexed.
            This callable can accept one (`value`) or two (`self`, `value`) arguments

        """

        attrs = {}
        if index_classes:
            attrs['index_classes'] = index_classes

        klass = type(str(name or cls.__name__), (cls, ), attrs)

        # let the ``configure`` method manage some fields
        configure_attrs = {}
        if key is not None:
            configure_attrs['key'] = key
        if transform is not None:
            configure_attrs['transform'] = transform

        if configure_attrs:
            klass = klass.configure(**configure_attrs)

        return klass

    @cached_property
    def _indexes(self):
        """Instantiate the indexes only when asked

        Returns
        -------
        list
            A list of all indexes, tied to the field.

        """
        return [index_class(field=self.field) for index_class in self.index_classes]

    def can_handle_suffix(self, suffix):
        """Tell if one of the managed indexes  can be used for the given filter prefix

        For parameters, see BaseIndex.can_handle_suffix

        """
        for index in self._indexes:
            if index.can_handle_suffix(suffix):
                return True

        return False

    def _reset_cache(self):
        """Reset attributes used to potentially rollback the indexes

        For the parameters, seen BaseIndex._reset_cache

        """
        for index in self._indexes:
            index._reset_cache()

    def _rollback(self):
        """Restore the index in its previous state

        For the parameters, seen BaseIndex._rollback

        """
        for index in self._indexes:
            index._rollback()

    def get_unique_index(self):
        """Returns the first index handling uniqueness

        Returns
        -------
        BaseIndex
            The first index capable of handling uniqueness

        Raises
        ------
        IndexError
            If not index is capable of handling uniqueness

        """
        return [index for index in self._indexes if index.handle_uniqueness][0]

    @property
    def handle_uniqueness(self):
        """Tell if at least one of the indexes can handle uniqueness

        Returns
        -------
        bool
            ``True`` if this multi-index can handle uniqueness.

        """
        try:
            self.get_unique_index()
        except IndexError:
            return False
        else:
            return True

    def prepare_args(self, args, transform=True):
        """Prepare args to be used by a sub-index

        Parameters
        ----------
        args: list
            The while list of arguments passed to add, check_uniqueness, get_filtered_keys...
        transform: bool
            If ``True``, the last entry in `args`, ie the value, will be transformed.
            Else it will be kept as is.

        """
        updated_args = list(args)
        if transform:
            updated_args[-1] = self.transform_value(updated_args[-1])
        if self.key:
            updated_args.insert(-1, self.key)

        return updated_args

    def check_uniqueness(self, *args):
        """For a unique index, check if the given args are not used twice

        For the parameters, seen BaseIndex.check_uniqueness

        """
        self.get_unique_index().check_uniqueness(*self.prepare_args(args, transform=False))

    def add(self, *args, **kwargs):
        """Add the instance tied to the field to all the indexes

        For the parameters, seen BaseIndex.add

        """

        check_uniqueness = kwargs.pop('check_uniqueness', False)
        args = self.prepare_args(args)

        for index in self._indexes:
            index.add(*args, check_uniqueness=check_uniqueness and index.handle_uniqueness, **kwargs)
            if check_uniqueness and index.handle_uniqueness:
                check_uniqueness = False

    def remove(self, *args):
        """Remove the instance tied to the field from all the indexes

        For the parameters, seen BaseIndex.remove

        """

        args = self.prepare_args(args)

        for index in self._indexes:
            index.remove(*args)

    def get_filtered_keys(self, suffix, *args, **kwargs):
        """Returns the index keys to be used by the collection for the given args

        For the parameters, see BaseIndex.get_filtered_keys

        """

        args = self.prepare_args(args, transform=False)

        for index in self._indexes:
            if index.can_handle_suffix(suffix):
                return index.get_filtered_keys(suffix, *args, **kwargs)

    def get_all_storage_keys(self):
        """Returns the keys to be removed by `clear` in aggressive mode

        For the parameters, see BaseIndex.get_all_storage_keys
        """

        keys = set()
        for index in self._indexes:
            keys.update(index.get_all_storage_keys())

        return keys


# This is a multi-indexes managing the different parts of a date in the format YYYY-MM-SS
DateIndexParts = MultiIndexes.compose([
    NumberRangeIndex.configure(prefix='year', transform=lambda value: value[:4], handle_uniqueness=False, name='YearIndex'),
    NumberRangeIndex.configure(prefix='month', transform=lambda value: value[5:7], handle_uniqueness=False, name='MonthIndex'),
    NumberRangeIndex.configure(prefix='day', transform=lambda value: value[8:10], handle_uniqueness=False, name='DayIndex'),
], name='DateIndexParts')

# A simple TextRangeIndex to filter on a date  in the format YYYY-MM-SS
DateRangeIndex = TextRangeIndex.configure(key='date', transform=lambda value: value[:10], name='DateRangeIndex')

# A full usable index for fields holding dates (without time)
DateIndex = MultiIndexes.compose([DateRangeIndex, DateIndexParts], name='DateIndex')

# This is a multi-indexes managing the different parts of a tine in the format HH:MM:SS
TimeIndexParts = MultiIndexes.compose([
    NumberRangeIndex.configure(prefix='hour', transform=lambda value: value[0:2], handle_uniqueness=False, name='HourIndex'),
    NumberRangeIndex.configure(prefix='minute', transform=lambda value: value[3:5], handle_uniqueness=False, name='MinuteIndex'),
    NumberRangeIndex.configure(prefix='second', transform=lambda value: value[6:8], handle_uniqueness=False, name='SecondIndex'),
], name='TimeIndexParts')

# A simple TextRangeIndex to filter on a date  in the format HH:MM:SS
TimeRangeIndex = TextRangeIndex.configure(key='time', transform=lambda value: value[:8], name='TimeRangeIndex')

# A full usable index for fields holding times (without date)
TimeIndex = MultiIndexes.compose([TimeRangeIndex, TimeIndexParts], name='TimeIndex')

# A full usable index for fields holding dates+times, without filtering on hour/min/sec
# but only full field, full date and full time, and year, month, day
DateSimpleTimeIndex = MultiIndexes.compose([
    TextRangeIndex.configure(key='full', name='FullDateTimeRangeIndex'),
    DateRangeIndex.configure(prefix='date'),
    DateIndexParts,
    TimeRangeIndex.configure(prefix='time', transform=lambda value: value[11:])  # pass only time
], name='DateSimpleTimeIndex', transform=lambda value: value[:19])

# A full usable index for fields holding dates+times, with full filtering capabilities
DateTimeIndex = MultiIndexes.compose([
    DateSimpleTimeIndex,
    TimeIndexParts.configure(transform=lambda value: value[11:]),
], name='DateTimeIndex')

# And a simple datetime index without parts
SimpleDateTimeIndex = MultiIndexes.compose([
    TextRangeIndex.configure(key='full', name='FullDateTimeRangeIndex'),
    DateRangeIndex.configure(prefix='date'),
    TimeRangeIndex.configure(prefix='time', transform=lambda value: value[11:])  # pass only time
], name='SimpleDateTimeIndex', transform=lambda value: value[:19])
